# coding=utf-8
""" Contains objects related to polygons.

"""

# import logging

from itertools import islice, combinations
from more_itertools import pairwise

from shapely import speedups
from shapely.coords import CoordinateSequence
from shapely.geometry import Polygon as ShapelyPolygon, MultiPolygon, Point as ShapelyPoint, LineString, MultiLineString
from shapely.ops import linemerge, unary_union, polygonize
from shapely.wkt import loads as wkt_to_geom
from osgeo.ogr import CreateGeometryFromWkt as wkt_to_ogr

from src.oogeo.objects.multipointbase import MultipointBase
import src.oogeo.objects.line
import src.oogeo.objects.point

def parse_multipolygon(point_list):
    """

    """
    assert isinstance(point_list, list), 'Points must be a list'

    if not isinstance(point_list[0], list):
        return MultiPolygon([parse_polygon(point_list)])
    else:
        return MultiPolygon([parse_polygon(poly_points) for poly_points in point_list])

def parse_polygon(point_list):
    """

    """
    assert isinstance(point_list, list), 'Points must be a list'

    exterior = []
    rings = []
    for point in point_list:
        if isinstance(point, tuple):
            exterior.append((point[0], point[1]))
        elif isinstance(point, dict):
            exterior.append((point['x'], point['y']))
        elif isinstance(point, src.oogeo.objects.point.Point):
            exterior.append((point.x, point.y))
        elif isinstance(point, list):
            ring = []
            for ring_point in point:
                if isinstance(ring_point, tuple):
                    ring.append((ring_point[0], ring_point[1]))
                elif isinstance(ring_point, dict):
                    ring.append((ring_point['x'], ring_point['y']))
                elif isinstance(point, src.oogeo.objects.point.Point):
                    ring.append((ring_point.x, ring_point.y))
            rings.append(ring)

    return ShapelyPolygon(exterior, rings)


def remove_invalid_vertices(shapely_polygon, spatial_reference=2263):
    """

    :param shapely_polygon:
    :param spatial_reference:
    :return:
    """

    # remove duplicate vertices
    env_polygon = Polygon(shapely_polygon, spatial_reference, allow_invalid=True)
    clean_env_polygon = env_polygon.remove_duplicate_vertices(allow_invalid=True)
    shapely_polygon = clean_env_polygon.geometry

    # if shapely_polygon.is_valid:
    #     return shapely_polygon, False

    # break the polygon down to edges
    duplicate_points = []
    polygons = []

    for polygon_index, polygon in enumerate(shapely_polygon):
        polygon_edges = []
        for ring_index, ring in enumerate([polygon.exterior] + [interior for interior in polygon.interiors]):
            ring_edges = []
            for point_1, point_2 in pairwise(list(ring.coords)):
                point_1 = src.oogeo.objects.point.Point(point_1, spatial_reference)
                point_2 = src.oogeo.objects.point.Point(point_2, spatial_reference)
                if point_1.equals(point_2):
                    duplicate_points.append({'polygon_index': polygon_index, 'ring_index': ring_index, 'point': point_2})
                    continue
                ring_edges.append(src.oogeo.objects.line.Line([point_1, point_2], spatial_reference))
            if len(ring_edges) > 0:
                polygon_edges.append(ring_edges)
        if len(polygon_edges) > 0:
            polygons.append(polygon_edges)

    # if we didn't find any valid polygons in the multipolygon, return none
    if len(polygons) == 0:
        return None, False

    # compare edges and find removal points
    removal_points = []
    for polygon_index, polygon_edges in enumerate(polygons):
        for ring_index, ring_edges in enumerate(polygon_edges):
            for edge_1, edge_2 in pairwise(ring_edges):
                angle_1 = edge_1.endpoints_angle + 360 if edge_1.endpoints_angle < 0 else edge_1.endpoints_angle
                angle_2 = edge_2.endpoints_angle + 360 if edge_2.endpoints_angle < 0 else edge_2.endpoints_angle
                angle_diff = abs(angle_1 - angle_2) - 180
                if -0.5 < angle_diff < 0.5:
                    removal_points.append({'polygon_index': polygon_index, 'ring_index': ring_index, 'point': edge_2.first_point})
            # finally, check the last edge against the first edge
            edge_1 = ring_edges[-1]
            edge_2 = ring_edges[0]
            angle_1 = edge_1.endpoints_angle + 360 if edge_1.endpoints_angle < 0 else edge_1.endpoints_angle
            angle_2 = edge_2.endpoints_angle + 360 if edge_2.endpoints_angle < 0 else edge_2.endpoints_angle
            angle_diff = abs(angle_1 - angle_2) - 180
            if -0.1 < angle_diff < 0.1:
                removal_points.append({'polygon_index': polygon_index, 'ring_index': ring_index, 'point': edge_2.first_point})

    # look for panhandles
    for polygon_index, polygon_edges in enumerate(polygons):
        for ring_index, ring_edges in enumerate(polygon_edges):
            if len(ring_edges) <= 4:
                continue
            # grab the angle of the last edge for initial comparison
            first_angle = ring_edges[-1].endpoints_angle + 360 if ring_edges[-1].endpoints_angle < 0 else ring_edges[-1].endpoints_angle
            idx = 0
            for middle_edge, last_edge in pairwise(ring_edges):
                middle_angle = middle_edge.endpoints_angle + 360 if middle_edge.endpoints_angle < 0 else middle_edge.endpoints_angle
                last_angle = last_edge.endpoints_angle + 360 if last_edge.endpoints_angle < 0 else last_edge.endpoints_angle
                angle_diff = abs(first_angle - last_angle) - 180
                if -5 < angle_diff < 5 and middle_edge.length < 0.1:
                    removal_points.append({'polygon_index': polygon_index, 'ring_index': ring_index, 'point': middle_edge.first_point})
                    removal_points.append({'polygon_index': polygon_index, 'ring_index': ring_index, 'point': middle_edge.last_point})
                first_angle = middle_angle
                idx += 1

    # reconstruct polygon and remove offensive points
    new_polygons = []
    adjusted = False
    for polygon_index, polygon in enumerate(shapely_polygon):
        new_rings = []
        for ring_index, ring in enumerate([polygon.exterior] + [interior for interior in polygon.interiors]):
            ring_points = []
            for point in ring.coords:
                point = src.oogeo.objects.point.Point(point, spatial_reference)
                remove = False
                for removal_point in removal_points:
                    if polygon_index == removal_point['polygon_index'] and ring_index == removal_point['ring_index'] and point.equals(removal_point['point']):
                        remove = True
                        break
                if not remove:
                    for duplicate_point in duplicate_points:
                        if polygon_index == duplicate_point['polygon_index'] and ring_index == duplicate_point['ring_index'] and point.equals(duplicate_point['point']):
                            duplicate_points.remove(duplicate_point)
                            remove = True
                            break
                if not remove:
                    ring_points.append(point.geometry)
                else:
                    adjusted = True
            if len(ring_points) >= 3:
                new_rings.append(LineString(ring_points))
        if len(new_rings) == 0:
            continue
        exterior_ring = new_rings[0]
        interior_rings = new_rings[1:]

        new_polygons.append(ShapelyPolygon(exterior_ring, interior_rings))

    if len(new_polygons) == 0:
        return None, adjusted
    return MultiPolygon(new_polygons), adjusted


def correct_invalid_polygon(shapely_polygon, spatial_reference=2263):
    """

    :param shapely_polygon:
    :return:
    """

    env_polygon = Polygon(shapely_polygon, spatial_reference, allow_invalid=True)
    env_polygon = env_polygon.remove_shards(allow_invalid=True)
    return env_polygon.geometry

    # if shapely_polygon.is_valid:
    #     return shapely_polygon
    # if not shapely_polygon.is_valid:
    #     zero_buffer_polygon = shapely_polygon.buffer(0)
    #     if isinstance(zero_buffer_polygon, ShapelyPolygon):
    #         zero_buffer_polygon = MultiPolygon([zero_buffer_polygon])
    #     shapely_polygon = zero_buffer_polygon if zero_buffer_polygon.is_valid and not zero_buffer_polygon.is_empty and abs(shapely_polygon.area - zero_buffer_polygon.area) < 2 else shapely_polygon
    #
    # for i in range(1, 5):
    #     shapely_polygon, adjusted = remove_invalid_vertices(shapely_polygon, spatial_reference)
    #     if shapely_polygon is None or not adjusted:
    #         if shapely_polygon is None:
    #             logging.info('  Invalid polygon could not be corrected - returning null')
    #         elif shapely_polygon.is_valid and adjusted:
    #             logging.info('  Invalid polygon was corrected - returning corrected polygon')
    #         elif not shapely_polygon.is_valid and not adjusted:
    #             # logging.info('  Invalid polygon could not be adjusted - wkt: {}'.format(shapely_polygon.wkt))
    #             logging.info('  Invalid polygon could not be adjusted')
    #         return shapely_polygon
    #
    # # logging.info('  Invalid polygon could not be adjusted after max number of attempts - wkt: {}'.format(shapely_polygon.wkt))
    # logging.info(
    #     '  Invalid polygon could not be adjusted after max number of attempts')
    # return shapely_polygon


class Polygon(MultipointBase):
    """ Provides support for analyzing polygons features in GIS layers.

    """

    def __init__(self, geometry, spatial_reference, allow_invalid=False):
        """ Initializes the polygon object.

        :param geometry: Geometry object representing a polygon. May be a list of coordinates (as tuples
            or dicts), ogr Polygon or MultiPolygon, well-known text representation, or Shapely Polygon
            or MultiPolygon.
        :param spatial_reference: The spatial reference for the geometry
        """

        if isinstance(geometry, str):
            geometry = wkt_to_geom(geometry)
        elif isinstance(geometry, CoordinateSequence):
            geometry = ShapelyPolygon(geometry)
        elif isinstance(geometry, list):
            geometry = parse_multipolygon(geometry)
        elif isinstance(geometry, ShapelyPolygon) or isinstance(geometry, MultiPolygon):
            geometry = wkt_to_geom(geometry.wkt)

        assert isinstance(geometry, ShapelyPolygon) or isinstance(geometry, MultiPolygon), 'Geometry must be a polygon'

        # elevate the geometry to a multigeometry if necessary
        if isinstance(geometry, ShapelyPolygon):
            geometry = MultiPolygon([geometry])

        super(Polygon, self).__init__(geometry, spatial_reference, allow_invalid)
        self._cached_points = None
        self.speedups_disabled = False

    @property
    def boundary(self):
        """ Returns the boundary of the polygon as a `Line()` object. If the polygon has an interior ring,
        it will be removed and only the exterior edge of the polygon will be returned.

        :return: A `Line()` object representing the boundary of the polygon.
        """
        return src.oogeo.objects.line.Line(self.geometry.boundary, self.spatial_reference, allow_invalid=self.allow_invalid)

    @property
    def area(self):
        """ Returns the area of the polygon.

        :return: A float value indicating the area of the polygon.
        """
        return self.geometry.area

    @property
    def mbr(self):
        """ Returns a `Polygon()` object representing the minimum bounding rectangle surrounding the polygon.
        The mbr rectangle is not oriented to the spatial axis and does not represent the *extent* of the geometry
        in XY coordinates.

        :return: A `Polygon()` object representing the minimum bounding rectangle surrounding the polygon.
        """
        mbr = wkt_to_geom(self.geometry.minimum_rotated_rectangle.wkt)
        if isinstance(mbr, ShapelyPolygon) or isinstance(mbr, MultiPolygon):
            if not mbr.is_valid:
                return None
            return Polygon(mbr, self.spatial_reference)
        if isinstance(mbr, LineString) or isinstance(mbr, MultiLineString):
            return src.oogeo.objects.line.Line(mbr, self.spatial_reference)
        if isinstance(mbr, ShapelyPoint):
            return src.oogeo.objects.point.Point(mbr, self.spatial_reference)

    @property
    def mbr_short_edges(self):
        """ Returns a `Line()` object representing the shortest edge of the minimum bounding rectangle surrounding
        the polygon. The mbr rectangle is not oriented to the spatial axis and does not represent the *extent* of
        the geometry in XY coordinates.

        :return: A `Line()` object representing the shortest edge of the minimum bounding rectangle
            surrounding the polygon.
        """
        mbr = self.mbr
        if not mbr or not isinstance(mbr, Polygon):
            return None
        edges = mbr.to_lines()
        short_edge = None
        for edge in edges:
            short_edge = edge if not short_edge or edge.length < short_edge.length else short_edge
        short_edges = []
        for edge in edges:
            if abs(short_edge.length - edge.length) < 0.1:
                short_edges.append(edge)

        return short_edges

    @property
    def mbr_long_edges(self):
        """ Returns a `Line()` object representing the longest edge of the minimum bounding rectangle surrounding
        the polygon. The mbr rectangle is not oriented to the spatial axis and does not represent the *extent* of
        the geometry in XY coordinates.

        :return: A `Line()` object representing the longest edge of the minimum bounding rectangle
            surrounding the polygon.
        """
        mbr = self.mbr
        if not mbr or not isinstance(mbr, Polygon):
            return None
        edges = mbr.to_lines()
        long_edge = None
        for edge in edges:
            long_edge = edge if not long_edge or edge.length > long_edge.length else long_edge

        long_edges = []
        for edge in edges:
            if abs(long_edge.length - edge.length) < 0.1:
                long_edges.append(edge)
        return long_edges

    def is_sliver(self, threshold):
        """ Indicates if the polygons is a sliver that is narrower than the provided threshold.

        :param threshold: The width below which a polygon is considered to be a sliver.
        :return: True if the width of the polygon is less than the threshold, otherwise False.
        """
        mbr_short_edges = self.mbr_short_edges
        if mbr_short_edges:
            return mbr_short_edges[0].length <= threshold
        # if no short edges come back this is a tiny tiny polygon, so it's a sliver by definition
        return False

    def is_panhandle(self, threshold):
        """ Indicates if the polygon has a panhandle that is narrower than the specified threshold.

        :param threshold: The width below which an area of the polygon is considered to be a panhandle.
        :return: True of the polygon contains a sliver, otherwise False.
        """
        lines = self.to_lines()
        lines.append(lines[0])
        lines.append(lines[1])
        first_line = lines[-2]
        middle_line = lines[-1]
        for last_line in lines:
            if src.oogeo.objects.line.Line.is_panhandle(first_line, middle_line, last_line, max_panhandle_width=threshold):
                return True
            first_line = middle_line
            middle_line = last_line
        return False

    def panhandle_width(self, threshold):
        """ The width of the panhandle if the polygon has a panhandle, or -1 if it does not.

        :param threshold: The width below which an area of the polygon is considered to be a panhandle.
        :return: The width of the panhandle, or -1 if the polygon does not have a panhandle.
        """
        lines = self.to_lines()
        lines.append(lines[0])
        lines.append(lines[1])
        first_line = lines[-2]
        middle_line = lines[-1]
        for last_line in lines:
            if src.oogeo.objects.line.Line.is_panhandle(first_line, middle_line, last_line, max_panhandle_width=threshold):
                distance_1 = first_line.midpoint.distance_to(last_line)
                distance_2 = last_line.midpoint.distance_to(first_line)
                return distance_1 if distance_1 < distance_2 else distance_2
            first_line = middle_line
            middle_line = last_line
        return -1

    def rings_as_polygons(self):
        """ Returns the interior rings of the polygon as polygons.

        :return: A list of `Polygon()` objects representing the interior rings of the polygon.
        """
        ring_polygons = []
        for polygon in self.geometry.geoms:
            for ring in polygon.interiors:
                ring_points = []
                for point in ring.coords:
                    ring_points.append(src.oogeo.objects.point.Point(point, self.spatial_reference))
                ring_polygons.append(Polygon(ring_points, self.spatial_reference))
        return ring_polygons

    @property
    def convex_hull(self):
        """ Returns a `Polygon()` object representing the convex hull surrounding the polygon.

        :return: A `Polygon()` object representing the convex hull surrounding the polygon.
        """
        return Polygon(self.geometry.convex_hull, self.spatial_reference)

    @property
    def centroid(self):
        """ Returns the centroid of the polygon as a `Point()` object for use in locating distances.

        :return: A `Point()` object representing the centroid of the polygon.
        """

        return src.oogeo.objects.point.Point(self.geometry.centroid, self.spatial_reference)

    @property
    def representative_point(self):
        """ Returns a point that is guaranteed to be within the boundaries of the polygon. This is
        usually the centroid of the geometry except in cases where the centroid falls outside of
        the bounds of the polygon, such as "L" and "C"-shaped polygons.

        :return: A `Point()` object representing the point within the polygon.
        """
        if self.contains(self.centroid, exact=True):
            return self.centroid
        return src.oogeo.objects.point.Point(self.geometry.representative_point(), self.spatial_reference)

    def _get_cached_points(self):
        """ Caches the vertices of the geometry as points and insures that the points are only collected once.

        :return: A list of `Point()` objects representing the vertices of the geometry.
        """
        if self._cached_points is not None:
            return self._cached_points

        self._cached_points = []
        for polygon in self.geometry.geoms:
            for ring in [polygon.exterior] + [interior for interior in polygon.interiors]:
                for point in ring.coords:
                    self._cached_points.append(src.oogeo.objects.point.Point(point, self.spatial_reference))

        return self._cached_points

    def get_point(self, index):
        """ Returns coordinates for vertex at the provided index of this Line.

        :param index: The integer index of the desired vertex. Negative integers will count backward from the last
            vertex, with -1 returning the last point.
        :return: Tuple (length 2) with x-y coordinates.
        """
        return self.points[index]

    @property
    def point_count(self):
        """ Returns the number of points in the geometry.

        :return: The number of points in the geometry.
        """
        return len(self.points)

    @property
    def length(self):
        """ Returns the length of the boundary of the polygon.

        """
        return self.boundary.length

    @property
    def points(self):
        """ Returns a list of coordinates representing the vertices of the line.

        :return: A list of `Point()` objects representing the points in the line.
        """
        return self._get_cached_points()

    def has_duplicate_vertices(self, decimals=None, tolerance=None):
        """ Indicates if the polygon has duplicate vertices.

        :return: True if the polygon has duplicate vertices, otherwise False.
        """
        for polygon in self.geometry.geoms:
            for ring in [polygon.exterior] + [interior for interior in polygon.interiors]:
                previous_point = None
                for test_point in ring.coords:
                    shapely_point = ShapelyPoint(test_point)
                    if previous_point:
                        if decimals:
                            if previous_point.almost_equals(shapely_point, decimals):
                                return True
                        elif tolerance:
                            if previous_point.distance(shapely_point) < tolerance:
                                return True
                        elif previous_point.wkt == shapely_point.wkt:
                            return True
                    previous_point = shapely_point
        return False

    def remove_duplicate_vertices(self, decimals=None, allow_invalid=False, tolerance=None):
        """ Creates a polygon based on the current polygon but with duplicate vertices removed.

        :return: A `Polygon()` object representing the current polygon with duplicate vertices removed.
        """

        # create the new geometry
        shapely_polygons = []
        for polygon_index, polygon in enumerate(self.geometry.geoms):
            shapely_lines = []
            for ring_index, ring in enumerate([polygon.exterior] + [interior for interior in polygon.interiors]):
                ring_points = []
                previous_point = None
                for point_index, test_point in enumerate(ring.coords):
                    shapely_point = ShapelyPoint(test_point)
                    if not previous_point:
                        ring_points.append(shapely_point)
                        previous_point = shapely_point
                        continue
                    if decimals:
                        if not previous_point.almost_equals(shapely_point, decimals):
                            ring_points.append(shapely_point)
                            previous_point = shapely_point
                    elif tolerance:
                        if not previous_point.distance(shapely_point) < tolerance:
                            ring_points.append(shapely_point)
                            previous_point = shapely_point
                    elif previous_point.wkt != shapely_point.wkt:
                        ring_points.append(shapely_point)
                        previous_point = shapely_point
                # if the first and last vertex do not match, that means the first index needs to go
                if ring_points[0].wkt != ring_points[-1].wkt:
                    ring_points[-1] = ring_points[0]
                if len(ring_points) > 2:
                    shapely_lines.append(LineString(ring_points))
            shapely_polygons.append(ShapelyPolygon(shapely_lines[0], shapely_lines[1:]))
        return Polygon(MultiPolygon(shapely_polygons), self.spatial_reference, allow_invalid=allow_invalid)

    def remove_shards(self, allow_invalid=False):
        """ Creates a polygon based on the current polygon but with duplicate vertices removed.

        :return: A `Polygon()` object representing the current polygon with duplicate vertices removed.
        """

        # first remove dupes
        working_geometry = self.remove_duplicate_vertices(allow_invalid=allow_invalid)

        # now find edges
        working_edges = working_geometry.to_lines()
        working_edges.append(working_edges[0])

        # if there are not enough edges, do not attempt
        if len(working_edges) < 4:
            return self.clone()

        # loop through the lines pairwise and find panhandles
        remove_points = []
        for edge_1, edge_2 in pairwise(working_edges):
            edge_angle_1 = edge_1.endpoints_angle + 360 if edge_1.endpoints_angle < 0 else edge_1.endpoints_angle
            edge_angle_2 = edge_2.endpoints_angle + 360 if edge_2.endpoints_angle < 0 else edge_2.endpoints_angle
            angle_diff = abs(edge_angle_1 - edge_angle_2) - 180
            if -0.5 < angle_diff < 0.5:
                remove_points.append(edge_1.last_point)

        for remove_point in remove_points:
            # logging.info('    Shard polygon found: %s', working_geometry.wkt)
            # logging.info('    Removing shard point: %s', remove_point.wkt)
            working_geometry = working_geometry.remove_point(remove_point)
            # logging.info('    Corrected polygon found: %s', working_geometry.wkt)
        return working_geometry

    def get_exterior_polygon(self):
        """ Returns the exterior (outer ring) of the polygon.

        :return: A `Polygon()` object representing the outer ring of the polygon.
        """
        return Polygon(self.geometry.exterior, self.spatial_reference)

    def to_lines(self):
        """ Converts the polygon to a set of `Line()` objects based on the polygon's vertices.

        :return: A list of `Line()` objects representing the edges of the polygon as lines.
        """
        lines = []
        for polygon in self.geometry.geoms:
            vertices = polygon.exterior.coords
            for v_a, v_b in zip(vertices[:-1], vertices[1:]):
                if not v_a == v_b:
                    lines.append(src.oogeo.objects.line.Line([v_a, v_b], self.spatial_reference))
            for interior in polygon.interiors:
                vertices = interior.coords
                for v_a, v_b in zip(vertices[:-1], vertices[1:]):
                    if not v_a == v_b:
                        lines.append(src.oogeo.objects.line.Line([v_a, v_b], self.spatial_reference))

        return lines

    def generalize(self, max_offset=1):
        """ Reduces the number of vertices in the polygon.

        Generalizing a line alters its path and will cause it to deviate from the original line depending on
        how many vertices are removed. The `max_offset` parameter specifies a maximum distance (in map units)
        that the line will be permitted to deviate from the original path. As the `max_offset` value grows larger,
        there will be fewer vertices in the resulting line, and the resulting line will deviate more from the
        original. Under the hood, this uses Shapely's `simplify` method.

        :param max_offset: The maximum distance that the line will be permitted to deviate
            from the original path.
        :return: A `Polygon` object representing the simplified polygon.
        """
        return Polygon(self.geometry.simplify(max_offset), self.spatial_reference)

    def union(self, other_polygon, allow_invalid=False):
        """ Merges two polygon objects together to form a single polygon.

        :param other_polygon: The other polygon object to be merged into this one.
        :return: A Polygon object representing the union of the two polygons.
        """
        return Polygon(unary_union([self.geometry, other_polygon.geometry]), self.spatial_reference, allow_invalid=allow_invalid)

    def cut(self, cutting_line):
        """ Cuts the polygon into two polygons along the provided cutting line.

        :param cutting_line: A `Line()` object representing the cutting edge.
        :return: A list with two `Polygon()` objects representing the two sides of the cut polygon.
        """
        cloned_geometry = self.clone()
        for cut_point in cutting_line.points:
            found_vertex = False
            for vertex in cloned_geometry.points:
                if cut_point.distance_to(vertex) <= 0.000001:
                    found_vertex = True
                    break
            if not found_vertex:
                cloned_geometry = cloned_geometry.add_point(cut_point)

        shapely_multi_geometry = cloned_geometry.geometry
        cut_geometry = cutting_line.geometry

        merged_edges = linemerge([shapely_multi_geometry.geoms[0].boundary, cut_geometry.geoms[0]])
        polygon_borders = unary_union(merged_edges)
        polygons = polygonize(polygon_borders)
        return_polygons = [Polygon(polygon, self.spatial_reference) for polygon in polygons]

        return return_polygons

    def remove_point(self, point, allow_invalid=False):
        """ Removes the specified vertex from the polygon.

        :param point: A `Point()` object representing the vertex to be removed.
        :return: A copy of this geometry without the specified vertex.
        """
        if not self.speedups_disabled:
            speedups.disable()
            self.speedups_disabled = True

        # create the new geometry
        shapely_polygons = []
        for polygon_index, polygon in enumerate(self.geometry.geoms):
            shapely_lines = []
            for ring_index, ring in enumerate([polygon.exterior] + [interior for interior in polygon.interiors]):
                ring_points = []
                first_point_removed = False
                last_point_removed = False
                for point_index, test_point in enumerate(ring.coords):
                    shapely_point = ShapelyPoint(test_point)
                    if shapely_point.wkt == point.geometry.wkt:
                        if point_index == 0:
                            first_point_removed = True
                        elif point_index == len(ring.coords) - 1:
                            last_point_removed = True
                        continue
                    ring_points.append(shapely_point)
                if first_point_removed and last_point_removed:
                    ring_points.append(ShapelyPoint(ring_points[0]))
                shapely_lines.append(LineString(ring_points))
            shapely_polygons.append(ShapelyPolygon(shapely_lines[0], shapely_lines[1:]))
        if not allow_invalid:
            for test_polygon in shapely_polygons:
                if not test_polygon.is_valid:
                    return self.clone()
        return Polygon(MultiPolygon(shapely_polygons), self.spatial_reference, allow_invalid=allow_invalid)

    def add_point(self, point, allow_invalid=False):
        """ Adds the specified point to the geometry.

        :param point: The point to be added.
        :return: A new geometry object with the new point inserted into the points list.
        """
        # if we already contain this point, return a clone of this geometry
        if self.contains_point(point):
            return self.clone()

        # find the nearest points to the specified point
        nearest_index = None
        for polygon_index, polygon in enumerate(self.geometry.geoms):
            for ring_index, ring in enumerate([polygon.exterior] + [interior for interior in polygon.interiors]):
                for point_index in range(len(ring.coords)):
                    # if this is the last vertex, associate it with the first one to make an edge
                    if point_index == len(ring.coords) - 1:
                        vertex_pair = [src.oogeo.objects.point.Point(ring.coords[point_index], self.spatial_reference),
                                       src.oogeo.objects.point.Point(ring.coords[0], self.spatial_reference)]
                    else:
                        vertex_pair = [src.oogeo.objects.point.Point(coords, self.spatial_reference) for coords in
                                       islice(ring.coords, point_index, point_index + 2)]
                    if vertex_pair[0].equals(vertex_pair[1]):
                        continue

                    # if this point equals the test point, do not add the point
                    # simply return a copy of the main geometry
                    if vertex_pair[0].equals(point):
                        return self.clone()
                    # if vertex_pair[0].distance_to(point) < 0.000001:
                    #     return Polygon(MultiPolygon(self.geometry), self.spatial_reference)

                    test_line = src.oogeo.objects.line.Line(vertex_pair, self.spatial_reference)
                    test_index = {
                        'polygon_index': polygon_index,
                        'ring_index': ring_index,
                        'point_index': point_index,
                        'distance': point.distance_to(test_line)
                    }
                    if not nearest_index or test_index['distance'] < nearest_index['distance']:
                        nearest_index = test_index

        # create the new geometry
        shapely_polygons = []
        for polygon_index, polygon in enumerate(self.geometry.geoms):
            shapely_lines = []
            for ring_index, ring in enumerate([polygon.exterior] + [interior for interior in polygon.interiors]):
                new_ring_points = []
                for point_index, test_point in enumerate(ring.coords):
                    new_ring_points.append(ShapelyPoint(test_point))
                    if nearest_index['polygon_index'] == polygon_index \
                        and nearest_index['ring_index'] == ring_index \
                        and nearest_index['point_index'] == point_index:
                        new_ring_points.append(ShapelyPoint(point.x, point.y))
                shapely_lines.append(LineString(new_ring_points))
            shapely_polygons.append(ShapelyPolygon(shapely_lines[0], shapely_lines[1:]))
        if not allow_invalid:
            for test_polygon in shapely_polygons:
                if not test_polygon.is_valid:
                    return self.clone()
        return Polygon(MultiPolygon(shapely_polygons), self.spatial_reference, allow_invalid=allow_invalid)

    def densify(self, distance):
        """ Add vertices to the polygon using the specified method.

        :param distance: The maximum distance between vertices. The actual distance between vertices will usually
            be less than the maximum distance as new vertices will be evenly distributed along the original segment.
        :return: A new `Polygon` object representing the densified polygon.
        """
        ogr_geometry = wkt_to_ogr(self.wkt)
        ogr_geometry.Segmentize(distance)
        return Polygon(wkt_to_geom(ogr_geometry.ExportToWkt()), self.spatial_reference)

    def snap_point(self, point):
        """ Returns the point along the edge of the polygon that is nearest to the indicated point.

        :param point: The point to be snapped.
        :return: `Point()` object representing a point along the line that is nearest to the specified point.
        """
        exterior_line = src.oogeo.objects.line.Line(self.geometry.geoms[0].exterior.coords, self.spatial_reference)
        distance_along_line = exterior_line.geometry.project(point.geometry)
        return exterior_line.point_at_distance(distance_along_line)

    def snap_to_vertex(self, other_geometry, tolerance, allow_invalid=False):
        """ Snaps the vertices of this geometry to the vertices of the provided geometry. Only points
        that fall within the specified tolerance will be snapped.

        :param other_geometry: The geometry to use as a snapping target.
        :param tolerance: The tolerance to use for snapping.
        :return: A copy of the original geometry with snapped vertices.
        """
        snap_points = other_geometry.points if isinstance(other_geometry, Polygon) or \
                      isinstance(other_geometry, src.oogeo.objects.line.Line) else [other_geometry]

        return_geometry = self.clone()

        for snap_point in snap_points:
            nearest_point = None
            distance = None
            for point in return_geometry.points:
                if not nearest_point:
                    nearest_point = point
                    distance = snap_point.distance_to(point)
                    continue
                test_distance = snap_point.distance_to(point)
                if test_distance < distance:
                    nearest_point = point
                    distance = test_distance

            if nearest_point and distance <= tolerance:
                return_geometry = return_geometry.replace_point(nearest_point, snap_point, allow_invalid=allow_invalid)
        return return_geometry

    def to_singlepart(self):
        """ Separates multi-part geometries into single parts.

        :return: A list of geometries representing the parts of the multi-part geometry.
        """

        if not self.is_multipart:
            return [self]

        return [Polygon(part, self.spatial_reference) for part in self.geometry.geoms]

    def symmetric_difference(self, other):
        """

        :return:
        """

        merged_self, merged_other = MultipointBase.merge_vertices(self, other)
        difference = merged_self.geometry.symmetric_difference(merged_other.geometry)
        if isinstance(difference, ShapelyPolygon) or isinstance(difference, MultiPolygon):
            if not difference.is_valid:
                difference = difference.buffer(0)
            return Polygon(difference, self.spatial_reference)
        return None

    def difference(self, other, exact=False, allow_invalid=False):
        """ Returns a `Polygon` object representing the difference between this polygon and the provided polygon.

        :param other: The other geometry to be compared.
        :param exact: If True, the geometries will be compared exactly as they are. If False, the vertices
            of the geometries will be merged prior to comparison to mitigate for tiny gaps and overlaps.
        :return: A `Polygon` object representing the difference between this polygon and the provided polygon.
        """
        if not exact:
            merged_self, merged_other = MultipointBase.merge_vertices(self, other)
        else:
            merged_self = self.clone()
            merged_other = other.clone()
        try:
            difference = merged_self.geometry.difference(merged_other.geometry)
            if isinstance(difference, ShapelyPolygon) or isinstance(difference, MultiPolygon):
                return Polygon(difference, self.spatial_reference, allow_invalid=allow_invalid) if difference is not None else None
            return None
        except:
            return None

    def clone(self):
        """ Creates a copy of the geometry.

        :return: A copy of the geometry.
        """
        return Polygon(MultiPolygon(self.geometry), self.spatial_reference, allow_invalid=self.allow_invalid)

    def snap_to_edge(self, other_geometry, tolerance, allow_invalid=False, densify=False):
        """ Snaps the vertices of this geometry to the vertices of the provided geometry. Only points
        that fall within the specified tolerance will be snapped.

        :param other_geometry: The geometry to use as a snapping target.
        :param tolerance: The tolerance to use for snapping.
        :return: A copy of the original geometry with snapped vertices.
        """
        other_edge = other_geometry.boundary if isinstance(other_geometry, Polygon) else other_geometry
        boundary = self.boundary

        # get the points in this geom that are close enough to the
        # target geom to be snapped
        local_points = list(filter(lambda x: other_edge.distance_to(x) <= tolerance, self.points))

        # also get the points in the target geom that are within range of this geom
        target_points = list(filter(lambda x: boundary.distance_to(x) <= tolerance, other_edge.points))

        # finally, get the edge segments that are near the area
        local_edges = list(filter(lambda x: other_edge.distance_to(x) <= tolerance, self.to_lines()))

        # find points to snap
        snap_points = []
        for local_point in local_points:
            # check local point to see if it is part of a perpendicular edge - if so don't snap it
            intersecting_edges = list(filter(lambda x: local_point.intersects(x), local_edges))
            parallel_edge_found = False
            for intersecting_edge in intersecting_edges:
                if intersecting_edge.is_parallel(other_edge):
                    parallel_edge_found = True
                    break

            snap_point = other_edge.snap_point(local_point)
            if len(target_points) > 0:
                nearest_target_point = list(sorted(target_points, key=lambda x: x.distance_to(local_point)))[0]
                # if the nearest target is close to the snap point, snap these points
                snap_diff = abs(local_point.distance_to(nearest_target_point) - local_point.distance_to(snap_point))
                if snap_diff < 1:
                    point_map = {'original_point': local_point.clone(), 'new_point': nearest_target_point,
                                 'snap_type': 'VERTEX_TO_VERTEX'}
                    snap_points.append(point_map)
                    continue

            # if we didn't find a parallel edge, move to the next point
            if not parallel_edge_found:
                continue

            # if there was not a vertex in the target that was close enough, find the nearest point on the edge
            point_map = {'original_point': local_point.clone(), 'new_point': snap_point,
                         'snap_type': 'VERTEX_TO_EDGE'}
            snap_points.append(point_map)

        # filter duplicate targets
        new_snap_points = []
        delete_points = []
        for snap_point in snap_points:
            found = False
            remove = False
            for new_snap_point in new_snap_points:
                if new_snap_point['new_point'].equals(snap_point['new_point']):
                    found = True
                    if not new_snap_point['original_point'].equals(snap_point['original_point']):
                        remove = True
                    break
            if not found:
                new_snap_points.append(snap_point)
            elif remove:
                delete_points.append(snap_point)

        # loop through and snap points
        snapped_geometry = self.clone()
        for snap_point in new_snap_points:
            snapped_geometry = snapped_geometry.replace_point(snap_point['original_point'], snap_point['new_point'], allow_invalid=allow_invalid)

        # remove unnecessary vertices
        for snap_point in delete_points:
            snapped_geometry = snapped_geometry.remove_point(snap_point['original_point'], allow_invalid=allow_invalid)

        # return here if we do not need to densify the edge
        if not densify:
            return snapped_geometry

        # find target points that need to be transferred
        missing_points = []
        for target_point in target_points:
            found = False
            for new_snap_point in new_snap_points:
                if new_snap_point['new_point'].equals(target_point):
                    found = True
                    break
            if not found:
                missing_points.append(target_point)

        for missing_point in missing_points:
            if snapped_geometry.boundary.distance_to(missing_point) < 0.1:
                snapped_geometry = snapped_geometry.add_point(missing_point, allow_invalid=allow_invalid)

        return snapped_geometry

    def replace_point(self, original_point, new_point, allow_invalid=False):
        """ Will replace the specified vertex with the new vertex. If the new vertex already exists in the
        geometry, the original vertex will be removed and no new vertex will be added.

        :param original_point: The original vertex to be replaced.
        :param new_point: The new vertex.
        :return: A geometry object with updated vertices.
        """

        # create the new geometry
        shapely_polygons = []
        replaced_point = False
        for polygon in self.geometry.geoms:
            shapely_lines = []
            for ring in [polygon.exterior] + [interior for interior in polygon.interiors]:
                new_ring_points = []
                for point in ring.coords:
                    shapely_point = ShapelyPoint(point)
                    if original_point.geometry.wkt == shapely_point.wkt:
                        new_ring_points.append(new_point.geometry)
                        replaced_point = True
                    else:
                        new_ring_points.append(shapely_point)
                shapely_lines.append(LineString(new_ring_points))
            shapely_polygons.append(ShapelyPolygon(shapely_lines[0], shapely_lines[1:]))
        if not replaced_point:
            # logging.warning('Could not replace point.')
            # logging.warning('  Original point wkt: %s', original_point.wkt)
            # logging.warning('  New point wkt: %s', new_point.wkt)
            return self.clone()
        return Polygon(MultiPolygon(shapely_polygons), self.spatial_reference, allow_invalid=allow_invalid)

    def replace_points(self, point_map):
        """ Will replace the specified vertex with the new vertex. If the new vertex already exists in the
        geometry, the original vertex will be removed and no new vertex will be added.

        :param original_point: The original vertex to be replaced.
        :param new_point: The new vertex.
        :return: A geometry object with updated vertices.
        """

        # remove any duplicate new points
        for map_1, map_2 in combinations(point_map, 2):
            if not map_1['new_point'] or not map_2['new_point']:
                continue
            if map_1['new_point'].distance_to(map_2['new_point']) < 0.5:
                if map_2['separation'] >= map_1['separation']:
                    map_2['new_point'] = None
                else:
                    map_1['new_point'] = None

        # check to see if we need to add the new point as a vertex
        for map in point_map:
            if not map['new_point']:
                continue
            for point in self.points:
                if point.equals(map['original_point']):
                    continue
                if not point.distance_to(map['new_point']) < 0.000001:
                    continue
                map['new_point'] = None
                break

        # loop through the geometry's points and add replace the point
        new_vertices = []

        for ring in [self.geometry.geoms[0].exterior] + [interior for interior in self.geometry.geoms[0].interiors]:
            new_ring = []
            for point in ring.coords:
                ring_point = src.oogeo.objects.point.Point(point, self.spatial_reference)
                for map in point_map:
                    if not ring_point.equals(map['original_point']):
                        continue
                    # if we reach here we have a vertex that needs to be replaced
                    ring_point = map['new_point']
                    break
                if ring_point:
                    new_ring.append(ring_point.clone())
            new_vertices.append(new_ring)

        if len(new_vertices[0]) < 3:
            # logging.info('  Could not replace point due to too few vertices. len(new_vertices[0]) = {}'.format(len(new_vertices[0])))
            return self.clone()
        test_polygon = Polygon(new_vertices, self.spatial_reference, allow_invalid=True)
        if test_polygon.is_valid:
            return test_polygon
        return self.clone()



"""
Unit tests for the Layer class. For more information see the
[API Tests](https://github.com/envelopecity/GIS/tree/dev#api-tests) section of the GIS repository.
"""

from src.oogeo.unittests.databasetest import DatabaseTest
from src.oogeo.objects.layer import Layer
from src.oogeo.objects.line import Line
from src.oogeo.objects.point import Point
from src.oogeo.objects.polygon import Polygon
from src.oogeo.objects.table import Table
from src.oogeo.objects.workspace import Workspace


class LayerClassUnitTests(DatabaseTest):
    def setUp(self):
        # load the test layer
        if not hasattr(self, 'test_layer'):
            self.workspace = Workspace(self._databases['testing_inputs'])
            self.test_layer = Layer('physical_blocks', self.workspace)
            self.overlay_layer = Layer('zoning_districts', self.workspace)
            self.adjacent_layer = Layer('adjacent_shapes', self.workspace)
            self.snap_targets_layer = Layer('snap_targets', self.workspace)

    def test_initialize_layer(self):
        # verify we can initialize a layer object without failure
        self.assertIsInstance(self.test_layer, Layer)

    def test_calculate_field(self):
        with self.test_layer.copy_to_new_table('layer_calc_field_test', output_schema='testing') as test_layer:
            test_point = Point((982609, 206224), self.test_layer.spatial_reference)
            test_layer.select_layer_by_location('CONTAINS', test_point)
            self.assertEqual(test_layer.count, 1)
            test_layer.calculate_field('borough', "'XX'")
            test_layer.select_layer_by_attribute("borough = 'XX'")
            self.assertEqual(test_layer.count, 1)
            test_layer.calculate_field('borough', "'MN'")

            # test shares a line segment with
            test_layer.select_layer_by_attribute(1)
            with test_layer.get_search_cursor([test_layer.shape]) as cursor:
                shared_polygon = next(cursor)[test_layer.shape]
            shared_line = shared_polygon.to_lines()[0]
            test_layer.select_layer_by_location('SHARE_A_LINE_SEGMENT_WITH', shared_line)
            self.assertEqual(test_layer.count, 1)
            test_layer.calculate_field('borough', "'XX'")
            test_layer.select_layer_by_attribute("borough = 'XX'")
            self.assertEqual(test_layer.count, 1)
            test_layer.calculate_field('borough', "'MN'")
            self.assertEqual(test_layer.count, 0)

            test_layer.select_layer_by_location('CONTAINS', test_point)
            test_layer.select_layer_by_location('SHARE_A_LINE_SEGMENT_WITH', shared_line, selection_type='ADD_TO_SELECTION')
            self.assertEqual(test_layer.get_selection_count(), 2)
            test_layer.calculate_field('borough', "'XX'")
            test_layer.select_layer_by_attribute("borough = 'XX'")
            self.assertEqual(test_layer.count, 2)
            test_layer.calculate_field('borough', "'MN'")
            self.assertEqual(test_layer.count, 0)

            # test cases for https://github.com/envelopecity/osgis/issues/158
            # add a few new polygons to the test layer
            left_polygon = Polygon([
                {'x': 0, 'y': 0},
                {'x': 2, 'y': 0},
                {'x': 2, 'y': 2},
                {'x': 0, 'y': 2},
                {'x': 0, 'y': 0},
            ], 2263)
            right_polygon = Polygon([
                {'x': 2, 'y': 0},
                {'x': 4, 'y': 0},
                {'x': 4, 'y': 2},
                {'x': 2, 'y': 2},
                {'x': 2, 'y': 0},
            ], 2263)

            test_layer.clear_selection()
            test_layer.add_rows([
                {'borough': 'XX', test_layer.shape: left_polygon},
                {'borough': 'XX', test_layer.shape: right_polygon}
            ])

            contained_polygon = Polygon([
                {'x': 0.5, 'y': 0.5},
                {'x': 1.5, 'y': 0.5},
                {'x': 1.5, 'y': 1.5},
                {'x': 0.5, 'y': 1.5},
                {'x': 0.5, 'y': 0.5},
            ], 2263)
            test_layer.select_layer_by_location('CONTAINS', contained_polygon)
            test_layer.calculate_field('borough', "'YY'")
            test_layer.select_layer_by_attribute("borough = 'YY'")
            self.assertEqual(test_layer.count, 1)
            test_layer.calculate_field('borough', "'XX'")

            crossing_polygon = Polygon([
                {'x': 1.5, 'y': 0.5},
                {'x': 2.5, 'y': 0.5},
                {'x': 2.5, 'y': 1.5},
                {'x': 1.5, 'y': 1.5},
                {'x': 1.5, 'y': 0.5},
            ], 2263)
            test_layer.select_layer_by_location('CONTAINS', crossing_polygon)
            test_layer.calculate_field('borough', "'YY'")
            test_layer.select_layer_by_attribute("borough = 'YY'")
            self.assertEqual(test_layer.count, 0)
            test_layer.calculate_field('borough', "'XX'")

            test_layer.select_layer_by_location('CROSSES', crossing_polygon)
            test_layer.calculate_field('borough', "'YY'")
            test_layer.select_layer_by_attribute("borough = 'YY'")
            self.assertEqual(test_layer.count, 2)
            test_layer.calculate_field('borough', "'XX'")

            test_layer.select_layer_by_attribute("borough = 'XX'")
            with self.workspace.create_feature_class('layer_calc_field_test_2', 'POLYGON', 2263, schema='testing_2') as query_layer:
                query_layer.add_field('attr_field')
                query_layer.add_rows([
                    {'attr_field': 'XX', query_layer.shape: crossing_polygon},
                    {'attr_field': 'XX', query_layer.shape: contained_polygon},
                ])
                self.assertEqual(query_layer.count, 2)
                query_layer.select_layer_by_location('WITHIN', test_layer)
                query_layer.calculate_field('attr_field', "'YY'")
                query_layer.select_layer_by_attribute("attr_field = 'YY'")
                self.assertEqual(query_layer.count, 1)

                # test setting string values w/out single quotes
                query_layer.select_layer_by_location('WITHIN', test_layer)
                query_layer.calculate_field('attr_field', "ZZ")
                query_layer.select_layer_by_attribute("attr_field = 'ZZ'")
                self.assertEqual(query_layer.count, 1)

                query_layer.select_layer_by_location('WITHIN', test_layer)
                query_layer.calculate_field('attr_field', "Z'Z")
                query_layer.select_layer_by_attribute("attr_field = 'Z''Z'")
                self.assertEqual(query_layer.count, 1)

                query_layer.select_layer_by_location('WITHIN', test_layer)
                query_layer.calculate_field('attr_field', "'Z'Z'")
                query_layer.select_layer_by_attribute("attr_field = 'Z''Z'")
                self.assertEqual(query_layer.count, 1)
        self.workspace.drop_schema('testing', cascade=True)
        self.workspace.drop_schema('testing_2', cascade=True)

    def test_select_layer_by_location(self):
        # create a point at a known location and select features there
        test_point = Point((982609, 206224), self.test_layer.spatial_reference)
        with self.test_layer.copy_to_new_table('select_test', output_schema='testing') as local_layer:
            local_layer.select_layer_by_location('CONTAINS', test_point)
            self.assertEqual(local_layer.get_selection_count(), 1)

            # create a test point outside of any shape and make sure nothing is selected
            test_point = Point((990126, 222502), local_layer.spatial_reference)
            local_layer.select_layer_by_location('CONTAINS', test_point)
            self.assertEqual(local_layer.get_selection_count(), 0)

            # create a polygon that contains exactly 2 features and test "WITHIN"
            within_polygon = Polygon([
                {'x': 989997, 'y': 222253},
                {'x': 990267, 'y': 222751},
                {'x': 991041, 'y': 222319},
                {'x': 990763, 'y': 221830},
            ], local_layer.spatial_reference)

            # create a polygon that crosses one block and envelopes another
            crossing_polygon = Polygon([
                {'x': 990203, 'y': 222156},
                {'x': 990469, 'y': 222648},
                {'x': 991041, 'y': 222316},
                {'x': 990773, 'y': 221837},
            ], local_layer.spatial_reference)

            # create a polygon that does not touch any others in the layer
            non_touching_polygon = Polygon([
                {'x': 986686, 'y': 206427},
                {'x': 986699, 'y': 206445},
                {'x': 986731, 'y': 206414},
            ], local_layer.spatial_reference)

            local_layer.select_layer_by_location('WITHIN', non_touching_polygon)
            self.assertEqual(local_layer.get_selection_count(), 0)

            local_layer.select_layer_by_location('CONTAINS', non_touching_polygon)
            self.assertEqual(local_layer.get_selection_count(), 0)

            local_layer.select_layer_by_location('WITHIN', within_polygon)
            self.assertEqual(local_layer.get_selection_count(), 2)

            # test additive functions
            test_point = Point((982609, 206224), local_layer.spatial_reference)
            local_layer.select_layer_by_location('CONTAINS', test_point, selection_type='ADD_TO_SELECTION')
            self.assertEqual(local_layer.get_selection_count(), 3)

            # test remove function
            local_layer.select_layer_by_location('WITHIN', within_polygon, selection_type='REMOVE_FROM_SELECTION')
            self.assertEqual(local_layer.get_selection_count(), 1)

            # test subset selection
            test_point = Point((990229, 222253), local_layer.spatial_reference)
            local_layer.select_layer_by_location('WITHIN', within_polygon)
            self.assertEqual(local_layer.get_selection_count(), 2)
            local_layer.select_layer_by_location('CONTAINS', test_point, selection_type='SUBSET_SELECTION')
            self.assertEqual(local_layer.get_selection_count(), 1)

            # test with crossing polygon
            local_layer.select_layer_by_location('WITHIN', crossing_polygon)
            self.assertEqual(local_layer.get_selection_count(), 0)

            # test with non-touching polygon
            local_layer.select_layer_by_location('WITHIN', non_touching_polygon)
            self.assertEqual(local_layer.get_selection_count(), 0)

            # test shares a line segment with
            local_layer.select_layer_by_attribute(1)
            with local_layer.get_search_cursor([local_layer.shape]) as cursor:
                shared_polygon = next(cursor)[local_layer.shape]
            shared_line = shared_polygon.to_lines()[0]
            local_layer.select_layer_by_location('SHARE_A_LINE_SEGMENT_WITH', shared_line)
            self.assertEqual(local_layer.get_selection_count(), 1)
            touching_line = Line([
                {'x': 987037.6, 'y': 206401.9},
                {'x': 986994.130654663, 'y': 206413.056383729},
            ], local_layer.spatial_reference)
            local_layer.select_layer_by_location('SHARE_A_LINE_SEGMENT_WITH', touching_line)
            self.assertEqual(local_layer.get_selection_count(), 0)

            # test boundary touches
            # local_layer.select_layer_by_location('BOUNDARY_TOUCHES', touching_line)
            # self.assertEqual(local_layer.get_selection_count(), 1)

            # test crossed by the outline of
            local_layer.select_layer_by_location('CROSSED_BY_THE_OUTLINE_OF', touching_line)
            self.assertEqual(local_layer.get_selection_count(), 0)
            crossing_line = Line([
                {'x': 986777.1, 'y': 206428.7},
                {'x': 986734.4, 'y': 206478.4},
            ], local_layer.spatial_reference)
            local_layer.select_layer_by_location('CROSSED_BY_THE_OUTLINE_OF', crossing_line)
            self.assertEqual(local_layer.get_selection_count(), 1)
            local_layer.select_layer_by_location('CROSSED_BY_THE_OUTLINE_OF', crossing_polygon)
            self.assertEqual(local_layer.get_selection_count(), 2)
            local_layer.select_layer_by_location('CROSSED_BY_THE_OUTLINE_OF', non_touching_polygon)
            self.assertEqual(local_layer.get_selection_count(), 0)

            # test have their centers in
            local_layer.select_layer_by_location('HAVE_THEIR_CENTER_IN', crossing_polygon)
            self.assertEqual(local_layer.get_selection_count(), 2)
            local_layer.select_layer_by_location('HAVE_THEIR_CENTER_IN', within_polygon)
            self.assertEqual(local_layer.get_selection_count(), 2)
            local_layer.select_layer_by_location('HAVE_THEIR_CENTER_IN', non_touching_polygon)
            self.assertEqual(local_layer.get_selection_count(), 0)

            # test within a distance
            local_layer.select_layer_by_location('WITHIN_A_DISTANCE', non_touching_polygon, search_distance=100)
            self.assertEqual(local_layer.get_selection_count(), 4)
            local_layer.select_layer_by_location('WITHIN_A_DISTANCE', non_touching_polygon, search_distance='30 METERS')
            self.assertEqual(local_layer.get_selection_count(), 4)

            # test layer selections
            local_layer.clear_selection()
            self.adjacent_layer.select_layer_by_location('WITHIN', local_layer)
            self.assertEqual(self.adjacent_layer.get_selection_count(), 6)
            self.adjacent_layer.clear_selection()

            local_layer.select_layer_by_location('CONTAINS', self.adjacent_layer)
            self.assertEqual(local_layer.get_selection_count(), 1)

            local_layer.select_layer_by_attribute(1)
            with local_layer.copy_to_new_table('single_shape_test') as single_shape_layer:
                local_layer.select_layer_by_location('SHARE_A_LINE_SEGMENT_WITH', single_shape_layer)
                self.assertEqual(local_layer.get_selection_count(), 1)
                local_layer.clear_selection()

            self.adjacent_layer.select_layer_by_location('HAVE_THEIR_CENTER_IN', local_layer)
            self.assertEqual(self.adjacent_layer.get_selection_count(), 6)
            self.adjacent_layer.clear_selection()

            local_layer.select_layer_by_location('WITHIN_A_DISTANCE', self.adjacent_layer, search_distance=200)
            self.assertEqual(local_layer.get_selection_count(), 9)

            local_layer.select_layer_by_location('INTERSECT', self.adjacent_layer, invert_spatial_relationship='INVERT')
            self.assertEqual(local_layer.get_selection_count(), 2878)

            # complex scenario #1
            # first search returns 4 records, the attribute search adds a fifth
            # and the within search adds two more, for a total of 7
            local_layer.select_layer_by_location('WITHIN_A_DISTANCE', non_touching_polygon, search_distance=100)
            local_layer.select_layer_by_attribute('{} = 1'.format(local_layer.oid), selection_type='ADD_TO_SELECTION')
            local_layer.select_layer_by_location('WITHIN', within_polygon, selection_type='ADD_TO_SELECTION')
            self.assertEqual(local_layer.get_selection_count(), 7)

            # remove 4 items from the previous selection
            local_layer.select_layer_by_location('WITHIN_A_DISTANCE', non_touching_polygon, search_distance=100, selection_type='REMOVE_FROM_SELECTION')
            self.assertEqual(local_layer.get_selection_count(), 3)

            # perform an attribute subselection
            local_layer.select_layer_by_attribute('{} = 1'.format(local_layer.oid), selection_type='SUBSET_SELECTION')
            self.assertEqual(local_layer.get_selection_count(), 1)

            # complex scenario #2
            local_layer.select_layer_by_attribute('{} = 1'.format(local_layer.oid))
            local_layer.select_layer_by_location('WITHIN_A_DISTANCE', non_touching_polygon, search_distance=100, selection_type='ADD_TO_SELECTION')
            local_layer.select_layer_by_location('WITHIN', within_polygon, selection_type='ADD_TO_SELECTION')
            self.assertEqual(local_layer.get_selection_count(), 7)
            local_layer.switch_selection()
            self.assertEqual(local_layer.get_selection_count(), 2872)

            # complex scenario #3
            local_layer.select_layer_by_attribute('{} = 1'.format(local_layer.oid))
            local_layer.select_layer_by_location('WITHIN_A_DISTANCE', non_touching_polygon, search_distance=100, selection_type='ADD_TO_SELECTION')
            local_layer.select_layer_by_attribute('{} = 2'.format(local_layer.oid), selection_type='ADD_TO_SELECTION')
            local_layer.select_layer_by_location('WITHIN', within_polygon, selection_type='ADD_TO_SELECTION')
            self.assertEqual(local_layer.get_selection_count(), 8)
            local_layer.select_layer_by_location('WITHIN', within_polygon, selection_type='REMOVE_FROM_SELECTION')
            self.assertEqual(local_layer.get_selection_count(), 6)
            local_layer.select_layer_by_attribute('{0} = 1 OR {0} = 2'.format(local_layer.oid), selection_type='SUBSET_SELECTION')
            self.assertEqual(local_layer.get_selection_count(), 2)
            local_layer.switch_selection()
            self.assertEqual(local_layer.get_selection_count(), 2877)
            local_layer.select_layer_by_location('WITHIN', within_polygon, selection_type='REMOVE_FROM_SELECTION')
            self.assertEqual(local_layer.get_selection_count(), 2875)
            local_layer.select_layer_by_attribute('{} = 2'.format(local_layer.oid), selection_type='ADD_TO_SELECTION')
            self.assertEqual(local_layer.get_selection_count(), 2876)
            local_layer.select_layer_by_attribute('{0} = 1 OR {0} = 2'.format(local_layer.oid), selection_type='SUBSET_SELECTION')
            self.assertEqual(local_layer.get_selection_count(), 1)

            # special share a line segment test
            with self.adjacent_layer.split_lines_to_layer('lines_selection_test_1') as line_layer_1, \
                line_layer_1.copy_to_new_table('lines_selection_test_2') as line_layer_2:
                line_layer_2.select_layer_by_location('SHARE_A_LINE_SEGMENT_WITH', line_layer_1)
                self.assertEqual(line_layer_1.count, line_layer_2.count)

                # use line layers to test MIDPOINT_NEAR option
                line_layer_2.select_layer_by_location('MIDPOINT_NEAR', line_layer_1)
                self.assertEqual(line_layer_1.count, line_layer_2.count)


            # test shares a line segment database function
            local_layer.select_layer_by_attribute(1)
            with local_layer.get_search_cursor([local_layer.shape]) as cursor:
                shared_polygon = next(cursor)[local_layer.shape]
            shared_line = shared_polygon.to_lines()[0]
            local_layer.select_layer_by_location('SHARE_A_LINE_SEGMENT_WITH', shared_line)
            self.assertEqual(local_layer.get_selection_count(), 1)
            touching_line = Line([
                {'x': 987037.6, 'y': 206401.9},
                {'x': 986994.130654663, 'y': 206413.056383729},
            ], local_layer.spatial_reference)
            local_layer.select_layer_by_location('SHARE_A_LINE_SEGMENT_WITH', touching_line)
            self.assertEqual(local_layer.get_selection_count(), 0)

            with self.adjacent_layer.split_lines_to_layer('lines_selection_test_1') as line_layer_1, \
                line_layer_1.copy_to_new_table('lines_selection_test_2') as line_layer_2:
                line_layer_2.select_layer_by_location('SHARE_A_LINE_SEGMENT_WITH', line_layer_1)
                self.assertEqual(line_layer_1.count, line_layer_2.count)

            local_layer.select_layer_by_attribute(1)
            with local_layer.copy_to_new_table('single_shape_test') as single_shape_layer:
                local_layer.select_layer_by_location('SHARE_A_LINE_SEGMENT_WITH', single_shape_layer)
                self.assertEqual(local_layer.get_selection_count(), 1)
                local_layer.clear_selection()

            # clear selections before exit
            local_layer.clear_selection()
            self.adjacent_layer.clear_selection()

        self.workspace.drop_schema('testing', cascade=True)

    def test_add_field(self):
        self.adjacent_layer.select_layer_by_attribute('{} <= 2'.format(self.adjacent_layer.oid))
        with self.adjacent_layer.copy_to_new_table('add_column_test', output_schema='testing') as test_layer:
            self.adjacent_layer.clear_selection()

            test_layer.add_field('CBBL_2')
            test_layer.copy_field_values('cbbl', 'CBBL_2')
            with test_layer.get_search_cursor(['cbbl', 'CBBL_2']) as cursor:
                for row in cursor:
                    self.assertEqual(row['cbbl'], row['CBBL_2'])
        self.workspace.drop_schema('testing', cascade=True)

    def test_duplicate_layer(self):
        # copy the test layer
        with self.test_layer.duplicate_layer('duplicated_layer_test') as duplicated_layer:
            # verify there are now 5 layers in the db
            layers = self.workspace.list_feature_classes()
            self.assertEqual(len(layers), 5)
            self.assertEqual(duplicated_layer.count, 0)
            for field_name in self.test_layer.field_names:
                self.assertIn(field_name, duplicated_layer.field_names)

    def test_delete_other_fields(self):
        with self.test_layer.copy_to_new_table('test_copy') as test_layer:
            test_layer.add_field('POINT_COUNT', 'SHORT')

            self.assertIn('POINT_COUNT', test_layer.field_names)

            # delete fields other than blocks and BLOCK_ID
            test_layer.delete_other_fields(['blocks', 'borough', 'phyblockid', 'cbb'])
            self.assertNotIn('POINT_COUNT', test_layer.field_names)
            self.assertIn('blocks', test_layer.field_names)
            self.assertIn('borough', test_layer.field_names)
            self.assertIn('phyblockid', test_layer.field_names)
            self.assertIn('cbb', test_layer.field_names)

    def test_shape_property(self):
        # test the shape property
        self.assertEqual(self.test_layer.shape, 'SHAPE')

    def test_oid_property(self):
        # test the oid property
        self.assertEqual(self.test_layer.oid, 'OBJECTID')

    def test_get_geometries(self):
        geometries = self.test_layer.get_geometries()
        self.assertIsInstance(geometries, list)
        self.assertGreater(len(geometries), 0)
        self.assertIsInstance(geometries[0], Polygon)

    def test_get_lines(self):
        self.test_layer.select_layer_by_attribute('OBJECTID < 11')
        lines = self.test_layer.get_lines()
        self.assertIsInstance(lines, list)
        self.assertGreater(len(lines), 0)
        self.assertIsInstance(lines[0], Line)
        self.test_layer.clear_selection()

    def test_field_names_property(self):
        self.assertIn('blocks', self.test_layer.field_names)
        self.assertIn('borough', self.test_layer.field_names)
        self.assertIn('phyblockid', self.test_layer.field_names)
        self.assertIn('cbb', self.test_layer.field_names)
        self.assertNotIn('OBJECTID', self.test_layer.field_names)
        self.assertNotIn('SHAPE', self.test_layer.field_names)

    def test_get_geodataframe(self):
        df = self.test_layer.get_geodataframe()
        self.assertEqual(len(df.index), 2879)
        self.assertEqual(df.index.name, self.test_layer.oid)

    def test_identity_to_layer(self):
        # select 1 block and one zd whose geometries overlap
        self.test_layer.select_layer_by_attribute('OBJECTID = 1092')
        self.overlay_layer.select_layer_by_attribute('OBJECTID = 327')
        with self.test_layer.identity_to_layer(self.overlay_layer, 'identity_test', output_schema='testing') as overlay_results_layer:

            # verify there is now 1 feature class in the schema
            layers = self.workspace.list_feature_classes(schema='testing')
            self.assertEqual(len(layers), 1)

            # verify that a layer object was returned and that it contains 2 features
            self.assertIsInstance(overlay_results_layer, Layer)
            self.assertEqual(overlay_results_layer.count, 2)
            self.assertIn('ZONEDIST', overlay_results_layer.field_names)
            overlay_results_layer.select_layer_by_attribute("ZONEDIST = 'C6-4.5'")
            self.assertEqual(overlay_results_layer.count, 1)
            overlay_results_layer.select_layer_by_attribute('ZONEDIST IS NULL')
            self.assertEqual(overlay_results_layer.count, 1)


        # select 1 block and 3 zds whose geometries overlap
        self.test_layer.select_layer_by_attribute('OBJECTID = 1092')
        self.overlay_layer.select_layer_by_attribute('OBJECTID IN (327, 465, 358)')
        with self.test_layer.copy_to_new_table('identity_inputs_test', output_schema='testing') as test_layer, \
                test_layer.identity_to_layer(self.overlay_layer, 'identity_test', output_schema='testing') as overlay_results_layer:

            # verify there are now 2 feature classes in the schema
            layers = self.workspace.list_feature_classes(schema='testing')
            self.assertEqual(len(layers), 2)

            # verify that a layer object was returned and that it contains 3 features
            self.assertIsInstance(overlay_results_layer, Layer)
            self.assertEqual(overlay_results_layer.count, 3)
            self.assertIn('ZONEDIST', overlay_results_layer.field_names)
            overlay_results_layer.select_layer_by_attribute("ZONEDIST = 'C6-4.5'")
            self.assertEqual(overlay_results_layer.count, 1)
            overlay_results_layer.select_layer_by_attribute("ZONEDIST = 'C6-6'")
            self.assertEqual(overlay_results_layer.count, 1)
            overlay_results_layer.select_layer_by_attribute("ZONEDIST = 'C5-3'")
            self.assertEqual(overlay_results_layer.count, 1)


        # clear selection and delete overlay results layer
        self.test_layer.clear_selection()
        self.overlay_layer.clear_selection()
        self.workspace.drop_schema('testing', cascade=True)

    def test_intersect_to_layer(self):
        # select 1 block and one zd whose geometries overlap
        self.test_layer.select_layer_by_attribute('OBJECTID = 1092')
        self.overlay_layer.select_layer_by_attribute('OBJECTID = 327')
        with self.test_layer.intersect_to_layer(self.overlay_layer, 'intersect_test', output_schema='testing') as overlay_results_layer:
            # verify there is now 1 layer in the db schema
            layers = self.workspace.list_feature_classes(schema='testing')
            self.assertEqual(len(layers), 1)

            # verify that a layer object was returned and that it contains  feature
            self.assertIsInstance(overlay_results_layer, Layer)
            self.assertEqual(overlay_results_layer.count, 1)
            self.assertEqual(overlay_results_layer.geometry_type, 'Polygon')
            self.assertIn('ZONEDIST', overlay_results_layer.field_names)
            with overlay_results_layer.get_search_cursor('ZONEDIST') as cursor:
                zonedist = next(cursor)['ZONEDIST']
                self.assertEqual(zonedist, 'C6-4.5')

        # with self.test_layer.intersect_to_layer(self.overlay_layer, 'intersect_point_test', output_type='POINT', output_schema='testing') as overlay_results_layer:
        #     self.assertIsInstance(overlay_results_layer, Layer)
        #     self.assertEqual(overlay_results_layer.geometry_type, 'Point')
        #     self.assertEqual(overlay_results_layer.count, 23)
        #     self.assertIn('ZONEDIST', overlay_results_layer.field_names)
        #
        # with self.overlay_layer.split_lines_to_layer('intersect_line_test', output_schema='testing_2') as line_layer, \
        #         self.test_layer.intersect_to_layer(line_layer, 'intersect_point_test_2', output_type='POINT', output_schema='testing') as overlay_results_layer:
        #     self.assertIsInstance(overlay_results_layer, Layer)
        #     self.assertEqual(overlay_results_layer.geometry_type, 'Point')
        #     self.assertEqual(overlay_results_layer.count, 2)

        # clear selection and delete overlay results layer
        self.test_layer.clear_selection()
        self.overlay_layer.clear_selection()
        self.workspace.drop_schema('testing', cascade=True)
        self.workspace.drop_schema('testing_2', cascade=True)

    def test_clip_to_layer(self):
        self.overlay_layer.select_layer_by_attribute('OBJECTID = 168')
        with self.overlay_layer.clip_to_layer(self.adjacent_layer, 'clipped_layer_test', output_schema='testing') as clip_layer:
            self.overlay_layer.clear_selection()
            layers = self.workspace.list_feature_classes()
            self.assertEqual(len(layers), 4)
            layers = self.workspace.list_feature_classes(schema='testing')
            self.assertEqual(len(layers), 1)
            self.assertIsInstance(clip_layer, Layer)
            self.assertEqual(clip_layer.count, 1)
        self.workspace.drop_schema('testing', cascade=True)

    def test_feature_vertices_to_points_layer(self):
        # select 1 block
        self.test_layer.select_layer_by_attribute('OBJECTID = 1092')
        with self.test_layer.feature_vertices_to_points_layer('vertices_to_points_test', output_schema='testing') as results_layer:

            # verify that a layer object was returned and that it contains 4 features
            self.assertIsInstance(results_layer, Layer)
            self.assertEqual(results_layer.count, 19)
            self.assertEqual(results_layer.geometry_type, 'Point')
            self.assertIn('blocks', results_layer.field_names)
            self.assertIn('borough', results_layer.field_names)
            self.assertIn('phyblockid', results_layer.field_names)
            self.assertIn('cbb', results_layer.field_names)

        # clear selection and delete overlay results layer
        self.test_layer.clear_selection()
        self.workspace.drop_schema('testing')

    def test_feature_to_point_layer(self):
        # select 1 block
        self.test_layer.select_layer_by_attribute('OBJECTID = 1092')
        with self.test_layer.feature_to_point_layer('feature_to_points_test', output_schema='testing') as results_layer:
            # verify that a layer object was returned and that it contains 1 feature
            self.assertIsInstance(results_layer, Layer)
            self.assertEqual(results_layer.count, 1)
            self.assertEqual(results_layer.geometry_type, 'Point')
            self.assertIn('blocks', results_layer.field_names)
            self.assertIn('borough', results_layer.field_names)
            self.assertIn('phyblockid', results_layer.field_names)
            self.assertIn('cbb', results_layer.field_names)

        with self.test_layer.feature_to_point_layer('feature_to_points_inside_test', point_location='INSIDE', output_schema='testing') as results_layer:
            # verify that a layer object was returned and that it contains 1 feature
            self.assertIsInstance(results_layer, Layer)
            self.assertEqual(results_layer.count, 1)
            self.assertEqual(results_layer.geometry_type, 'Point')
            self.assertIn('blocks', results_layer.field_names)
            self.assertIn('borough', results_layer.field_names)
            self.assertIn('phyblockid', results_layer.field_names)
            self.assertIn('cbb', results_layer.field_names)

        # clear selection and delete overlay results layer
        self.test_layer.clear_selection()
        self.workspace.drop_schema('testing')

    def test_copy_to_new_table(self):
        # select a single feature and create a copy of the polygon layer
        self.test_layer.select_layer_by_attribute(1)
        copy_layer = self.test_layer.copy_to_new_table('copy_layer_test')
        self.test_layer.clear_selection()

        # verify there are now 5 layers in the db and there is only one record in the output table
        layers = self.workspace.list_feature_classes()
        self.assertEqual(len(layers), 5)
        self.assertEqual(copy_layer.count, 1)

        # delete test layer
        copy_layer.delete()

    def test_select_to_layer(self):
        # create a layer from the polygon layer
        with self.test_layer.select_to_layer('copy_layer_test', 'OBJECTID = 1', output_schema='testing') as select_layer:

            # verify there are now 5 layers in the db
            layers = self.workspace.list_feature_classes()
            self.assertEqual(len(layers), 4)
            layers = self.workspace.list_feature_classes(schema='testing')
            self.assertEqual(len(layers), 1)

            self.assertEqual(select_layer.count, 1)

            # verify there is only one record in the new layer
            self.assertEqual(select_layer.get_selection_count(), 1)

        self.workspace.drop_schema('testing', cascade=True)

    def test_convert_to_table(self):
        with self.test_layer.convert_to_table('converted_table_test', output_schema='testing') as test_table:
            self.assertIsInstance(test_table, Table)

            tables = self.workspace.list_tables()
            layers = self.workspace.list_feature_classes()
            self.assertEqual(len(tables), 1)
            self.assertEqual(len(layers), 4)

            tables = self.workspace.list_tables(schema='testing')
            layers = self.workspace.list_feature_classes(schema='testing')
            self.assertEqual(len(tables), 1)
            self.assertEqual(len(layers), 0)

            self.assertIn(test_table.table_name, tables)
            self.assertNotIn(test_table.table_name, layers)
            self.assertEqual(self.test_layer.count, test_table.count)

        self.workspace.drop_schema('testing', cascade=True)

    def test_geometry_type(self):
        self.assertEqual(self.test_layer.geometry_type, 'Polygon')

    def test_spatial_reference(self):
        self.assertEqual(self.test_layer.spatial_reference, 2263)

    def test_get_geometry_columns(self):
        geometry_columns = self.test_layer.get_geometry_columns()
        self.assertEqual(len(geometry_columns), 1)
        self.assertEqual(geometry_columns[0]['name'], 'SHAPE')

        geometry_columns = self.test_layer.get_geometry_columns(include_hidden=True)
        self.assertEqual(len(geometry_columns), 2)
        names = [geom_col['name'] for geom_col in geometry_columns]
        self.assertIn('SHAPE', names)
        # self.assertIn('edge_buffers', names)
        self.assertIn('EDGE_BUFFERS', names)

    def test_mbr_to_layer(self):
        # create an mbr layer from the polygon layer
        with self.test_layer.minimum_bounding_geometry_to_layer('mbr_layer_test', output_schema='testing') as mbr_layer:
            layers = self.workspace.list_feature_classes()
            self.assertEqual(len(layers), 4)
            layers = self.workspace.list_feature_classes(schema='testing')
            self.assertEqual(len(layers), 1)

            # verify that there are the same number of rows as the
            # source table, and that there is an "orig_fid" field
            self.assertEqual(self.test_layer.count, mbr_layer.count)
            self.assertIn('orig_fid', mbr_layer.field_names)

        # create an mbr layer from the polygon layer using field options
        with self.test_layer.minimum_bounding_geometry_to_layer('mbr_layer_2_test',
            geometry_type='CONVEX_HULL',
            omit_field_options='MBG_FIELDS',
            output_schema='testing') as mbr_layer:

            layers = self.workspace.list_feature_classes()
            self.assertEqual(len(layers), 4)
            layers = self.workspace.list_feature_classes(schema='testing')
            self.assertEqual(len(layers), 1)

            # verify that there are the same number of rows as the
            # source table, and that there is an "orig_fid" field as well as the MBG fields
            self.assertEqual(self.test_layer.count, mbr_layer.count)
            self.assertIn('orig_fid', mbr_layer.field_names)
            self.assertIn('mbg_length', mbr_layer.field_names)
            self.assertIn('mbg_width', mbr_layer.field_names)

        self.workspace.drop_schema('testing', cascade=True)

    def test_buffer_to_layer(self):
        # create a buffer layer from the polygon layer
        with self.test_layer.buffer_to_layer('buffer_layer_test', 100, output_schema='testing') as buffer_layer:
            layers = self.workspace.list_feature_classes()
            self.assertEqual(len(layers), 4)
            layers = self.workspace.list_feature_classes(schema='testing')
            self.assertEqual(len(layers), 1)
        self.workspace.drop_schema('testing')

    def test_dissolve_to_layer(self):
        # dissolve the adjacent polygons layer by cbb
        with self.adjacent_layer.dissolve_to_layer('dissolve_layer_test', ['cbb'], output_schema='testing') as dissolve_layer:
            self.assertEqual(dissolve_layer.count, 1)
            self.assertNotIn('cbbl', dissolve_layer.field_names)
            self.assertIn('cbb', dissolve_layer.field_names)

        # dissolve the adjacent polygons layer by cbbl (should not dissolve)
        with self.adjacent_layer.dissolve_to_layer('dissolve_layer_2_test', ['cbbl'], output_schema='testing') as dissolve_layer_2:
            self.assertEqual(dissolve_layer_2.count, 6)
            self.assertIn('cbbl', dissolve_layer_2.field_names)
            self.assertNotIn('cbb', dissolve_layer_2.field_names)

        # test multi-part versus single-part
        with self.adjacent_layer.dissolve_to_layer('dissolve_layer_3_test', ['canbuild'], output_schema='testing') as dissolve_layer_3:
            self.assertEqual(dissolve_layer_3.count, 2)
            self.assertIn('canbuild', dissolve_layer_3.field_names)
            self.assertNotIn('cbb', dissolve_layer_3.field_names)

        with self.adjacent_layer.dissolve_to_layer('dissolve_layer_4_test', ['canbuild'], multi_part=False, output_schema='testing') as dissolve_layer_4:
            self.assertEqual(dissolve_layer_4.count, 3)
            self.assertIn('canbuild', dissolve_layer_4.field_names)
            self.assertNotIn('cbb', dissolve_layer_4.field_names)

        # test dissolving on multiple fields
        with self.adjacent_layer.dissolve_to_layer('dissolve_layer_5_test', ['canbuild', 'corner'], output_schema='testing') as dissolve_layer_5:
            self.assertEqual(dissolve_layer_5.count, 3)
            self.assertIn('canbuild', dissolve_layer_5.field_names)
            self.assertIn('corner', dissolve_layer_5.field_names)
            self.assertNotIn('cbb', dissolve_layer_5.field_names)

        with self.adjacent_layer.dissolve_to_layer('dissolve_layer_6_test', ['canbuild', 'corner'], multi_part=False, output_schema='testing') as dissolve_layer_6:
            self.assertEqual(dissolve_layer_6.count, 5)
            self.assertIn('canbuild', dissolve_layer_6.field_names)
            self.assertIn('corner', dissolve_layer_6.field_names)
            self.assertNotIn('cbb', dissolve_layer_6.field_names)

        # test dissolving line layers
        self.adjacent_layer.select_layer_by_attribute(1)
        with self.adjacent_layer.split_lines_to_layer('temp_lines_test', output_schema='testing_2') as line_layer:
            self.adjacent_layer.clear_selection()
            with line_layer.dissolve_to_layer('temp_lines_dissolved_test', output_schema='testing') as dissolve_layer_7:
                self.assertLess(dissolve_layer_7.count, line_layer.count)

        self.workspace.drop_schema('testing', cascade=True)
        self.workspace.drop_schema('testing_2', cascade=True)

    def test_split_and_unsplit_lines_to_layer(self):
        # export polygon edges to lines
        with self.adjacent_layer.split_lines_to_layer('split_lines_test', output_schema='testing') as lines_layer:
            self.assertEqual(lines_layer.geometry_type, 'Polyline')
            self.assertEqual(lines_layer.count, 45)

            # merge the lines back together
            with lines_layer.unsplit_lines_to_layer('unsplit_lines_test') as unsplit_lines:
                self.assertEqual(unsplit_lines.geometry_type, 'Polyline')
                self.assertEqual(unsplit_lines.count, 41)
                self.assertEqual(unsplit_lines.schema, 'testing')

            with lines_layer.unsplit_lines_to_layer('dissolve_lines_test', dissolve_fields='cbbl', output_schema='testing_2') as dissolve_lines:
                self.assertEqual(dissolve_lines.geometry_type, 'Polyline')
                self.assertEqual(dissolve_lines.count, 6)
                self.assertEqual(dissolve_lines.schema, 'testing_2')

        self.workspace.drop_schema('testing', cascade=True)
        self.workspace.drop_schema('testing_2', cascade=True)

    def test_near(self):
        # select the 9 blocks around the adjacent lots layer
        self.test_layer.select_layer_by_location('WITHIN_A_DISTANCE', self.adjacent_layer, search_distance=200)
        with self.test_layer.copy_to_new_table('near_layer_test', output_schema='testing') as near_layer:
            self.test_layer.clear_selection()

            # use the near function to measure the distances of the 9 selected polygons from the adjacent lots layer
            near_layer.near(self.adjacent_layer)

            # verify that all rows in the near table have a near_fid value
            near_layer.select_layer_by_attribute('near_fid >= 0')
            self.assertEqual(near_layer.count, 9)

            # verify that 8 records have a near distance and one has the value 0
            near_layer.select_layer_by_attribute('near_dist > 0')
            self.assertEqual(near_layer.count, 8)
            near_layer.select_layer_by_attribute('near_dist = 0')
            self.assertEqual(near_layer.count, 1)
            near_layer.clear_selection()

            # test with a selection set on the near layer
            near_layer.select_layer_by_location('INTERSECT', self.adjacent_layer)
            near_layer.switch_selection()

            # measure the distance to the adjacent lot polygons
            near_layer.near(self.adjacent_layer)

            # verify that all rows but one have a near_fid value (one row should be excluded
            near_layer.select_layer_by_attribute('near_fid >= 0')
            self.assertEqual(near_layer.count, 8)
            near_layer.select_layer_by_attribute('near_fid = -1')
            self.assertEqual(near_layer.count, 1)

            # verify that 8 records have a near distance and one has the value -1
            near_layer.select_layer_by_attribute('near_dist > 0')
            self.assertEqual(near_layer.count, 8)
            near_layer.select_layer_by_attribute('near_dist = -1')
            self.assertEqual(near_layer.count, 1)

            # test with a selection set on both layers
            near_layer.select_layer_by_location('INTERSECT', self.adjacent_layer)
            near_layer.switch_selection()
            self.adjacent_layer.select_layer_by_attribute('{} = 1'.format(self.adjacent_layer.oid))

            # measure the distance to the adjacent lot polygons
            near_layer.near(self.adjacent_layer)

            # verify that all rows but one have a near_fid value (one row should be excluded
            near_layer.select_layer_by_attribute('near_fid = 1')
            self.assertEqual(near_layer.count, 8)
            near_layer.select_layer_by_attribute('near_fid = -1')
            self.assertEqual(near_layer.count, 1)

            # verify that 8 records have a near distance and one has the value -1
            near_layer.select_layer_by_attribute('near_dist > 0')
            self.assertEqual(near_layer.count, 8)
            near_layer.select_layer_by_attribute('near_dist = -1')
            self.assertEqual(near_layer.count, 1)

            self.adjacent_layer.clear_selection()

        self.workspace.drop_schema('testing', cascade=True)

    def test_snap(self):
        # snap the vertices of the adjacent lots layer to the main test layer (physical blocks)
        with self.snap_targets_layer.copy_to_new_table('snap_targets_test', output_schema='testing') as temp_snap_targets:
            temp_snap_targets.snap([[self.test_layer, 'VERTEX', '1 FOOT']])
            self.test_layer.select_layer_by_location('INTERSECT', self.adjacent_layer)
            with self.test_layer.split_lines_to_layer('snap_lines_test', output_schema='testing_2') as snap_lines_layer:
                self.test_layer.clear_selection()
                temp_snap_targets.snap([[snap_lines_layer, 'EDGE', '2 FEET']])
                temp_snap_targets.snap([[snap_lines_layer, 'END', '4 FEET']])

            # compare the first shape in the adjacent and target layers - they should now almost match
            self.adjacent_layer.select_layer_by_attribute('{} = 1'.format(self.adjacent_layer.oid))
            with self.adjacent_layer.get_search_cursor([self.adjacent_layer.shape]) as cursor:
                adjacent_geometry = next(cursor)[self.adjacent_layer.shape]
            self.adjacent_layer.clear_selection()
            temp_snap_targets.select_layer_by_location('INTERSECT', adjacent_geometry.centroid)
            with temp_snap_targets.get_search_cursor([temp_snap_targets.shape]) as cursor:
                snapped_geometry = next(cursor)[temp_snap_targets.shape]
            temp_snap_targets.clear_selection()
            self.assertAlmostEqual(adjacent_geometry.area, snapped_geometry.area, 4)

        self.workspace.drop_schema('testing', cascade=True)
        self.workspace.drop_schema('testing_2', cascade=True)

    def test_split_line_at_points_to_layer(self):
        # export polygon edges to lines
        with self.adjacent_layer.split_lines_to_layer('lines_to_split_test', output_schema='testing') as lines_layer, \
            self.adjacent_layer.feature_vertices_to_points_layer('split_points_test', output_schema='testing_2') as points_layer:
            # merge the lines back together so we can split them with the points
            with lines_layer.unsplit_lines_to_layer('dissolve_lines_test') as dissolve_lines:
                self.assertEqual(dissolve_lines.schema, 'testing')
                # split the lines at points - should result in 45 lines
                with dissolve_lines.split_lines_at_points_to_layer(points_layer, 'lines_split_at_points_test', output_schema='testing_2') as split_lines:
                    self.assertEqual(split_lines.geometry_type, 'Polyline')
                    self.assertEqual(split_lines.count, 45)
                    self.assertEqual(split_lines.schema, 'testing_2')

        self.workspace.drop_schema('testing', cascade=True)
        self.workspace.drop_schema('testing_2', cascade=True)

    def test_merge_layers(self):
        # select a single feature and create a copy of the polygon layer
        self.test_layer.select_layer_by_attribute('{} = 1'.format(self.test_layer.oid))
        with self.test_layer.copy_to_new_table('copy_layer_test', output_schema='testing') as copy_layer:
            self.test_layer.clear_selection()

            # merge the copied layer into the test layer
            merged_layer = self.test_layer.merge_to_layer(copy_layer, 'merged_layer_test')
            self.assertEqual(self.test_layer.count, merged_layer.count - 1)
            self.assertEqual(merged_layer.schema, 'public')
            merged_layer.delete()

            merged_layer = self.test_layer.merge_to_layer(copy_layer, 'merged_layer_2_test', output_schema='testing')
            self.assertEqual(self.test_layer.count, merged_layer.count - 1)
            self.assertEqual(merged_layer.schema, 'testing')
            merged_layer.delete()

        self.workspace.drop_schema('testing', cascade=True)

    def test_merge_layers_to_layer(self):
        # select a single feature and create a copy of the polygon layer
        self.test_layer.select_layer_by_attribute('OBJECTID = 1')
        with self.test_layer.copy_to_new_table('merge_copy_layer_test', output_schema='testing') as copy_layer:
            self.test_layer.clear_selection()

            # merge the copied layer into the test layer
            merged_layer = Layer.merge_layers_to_layer([copy_layer, self.test_layer], 'merged_layers_to_layer_test', self.workspace)
            self.assertEqual(merged_layer.schema, 'public')
            self.assertEqual(self.test_layer.count, merged_layer.count - 1)
            merged_layer.delete()

            # merge the copied layer into the test layer
            merged_layer = Layer.merge_layers_to_layer([copy_layer, self.test_layer], 'merged_layers_to_layer_test',
                                                       self.workspace, output_schema='testing_2')

            self.assertEqual(merged_layer.schema, 'testing_2')
            self.assertEqual(self.test_layer.count, merged_layer.count - 1)
            merged_layer.delete()

        self.workspace.drop_schema('testing', cascade=True)
        self.workspace.drop_schema('testing_2', cascade=True)

    def test_union_geometries(self):
        # union the geometries in the adjacent layer
        geometry = self.adjacent_layer.union_geometries()
        self.assertIsInstance(geometry, Polygon)

    def test_single_part_layer(self):

        with self.test_layer.to_single_part_layer('single_part_layer_test', output_schema='testing') as single_part_layer:
            self.assertEqual(single_part_layer.geometry_type, 'Polygon')
            self.assertEqual(self.test_layer.count, single_part_layer.count)
            self.assertEqual(single_part_layer.schema, 'testing')

        self.workspace.drop_schema('testing', cascade=True)

    def test_multi_part_layer(self):

        with self.test_layer.to_multi_part_layer('multi_part_layer_test', 'borough', output_schema='testing') as multi_part_layer:
            self.assertEqual(multi_part_layer.geometry_type, 'Polygon')
            self.assertEqual(multi_part_layer.count, 1)
            self.assertEqual(multi_part_layer.schema, 'testing')

        self.workspace.drop_schema('testing', cascade=True)

    def test_feature_to_polygon_layer(self):
        # create a point layer from the polygon layer
        self.test_layer.select_layer_by_attribute('OBJECTID = 1092')
        with self.test_layer.split_lines_to_layer('lines_layer_test', output_schema='testing') as lines_layer:
            self.test_layer.clear_selection()
            with self.test_layer.feature_to_point_layer('label_centroids_test', output_schema='testing_2') as points_layer:
                points_layer.delete_other_fields(['cbb'])

                # re-merge lines together as polygons
                with lines_layer.feature_to_polygon_layer('remade_polygon_layer_test', attributes_layer=points_layer, output_schema='testing_3') as polygon_layer:
                    # verify that the output is a polygon layer
                    self.assertEqual(polygon_layer.geometry_type, 'Polygon')
        self.workspace.drop_schema('testing', cascade=True)
        self.workspace.drop_schema('testing_2', cascade=True)
        self.workspace.drop_schema('testing_3', cascade=True)

    def test_generate_adjacency_table(self):
        # generate adjacency table from layer
        with self.adjacent_layer.generate_adjacency_table('adjacency_table_test', id_field='cbbl', output_schema='testing') as adjacency_table:
            # verify there is now 1 layer and 2 tables in the db and that adjacency table is of type "Table"
            tables = self.workspace.list_tables(schema='testing')
            self.assertEqual(len(tables), 1)
            self.assertIn('adjacency_table_test', tables)
            self.assertIsInstance(adjacency_table, Table)
            self.assertEqual(adjacency_table.count, 16)

        self.workspace.drop_schema('testing', cascade=True)

    def test_copy_to_schema(self):
        output_layer = self.test_layer.copy_to_new_table(self.test_layer.table_name, output_schema='testing')
        self.assertIsInstance(output_layer, Layer)
        layer_list = self.workspace.list_tables_and_layers(schema='testing')
        self.assertEqual(len(layer_list), 1)
        self.assertIn('physical_blocks', layer_list)
        output_layer.delete()
        layer_list = self.workspace.list_tables_and_layers(schema='testing')
        self.assertEqual(len(layer_list), 0)
        self.workspace.drop_schema('testing')

    def test_reproject_to_layer(self):
        self.test_layer.select_layer_by_attribute('OBJECTID = 1092')
        with self.test_layer.reproject_to_layer('reproject_test', 4326) as reprojected_layer:
            self.assertEqual(reprojected_layer.count, 1)
            self.assertEqual(reprojected_layer.spatial_reference, 4326)
            with reprojected_layer.get_search_cursor(reprojected_layer.shape) as cursor:
                geometry = next(cursor)[reprojected_layer.shape]
                self.assertEqual(geometry.spatial_reference, 4326)
                self.assertEqual(geometry.wkt, 'MULTIPOLYGON (((-73.98480133968964 40.74839358146502, -73.98514584121077 40.74791901190429, -73.98672799859544 40.748579010476696, -73.98680181748703 40.748609824374135, -73.98701632340045 40.74869932394527, -73.9870826384487 40.74872702556731, -73.98728315983132 40.74881069409462, -73.98742412455863 40.7488694865071, -73.98789395158482 40.7490652636495, -73.98782153208903 40.749457490345264, -73.98771225610628 40.749608230654204, -73.98707955145927 40.74934409792781, -73.98700047942175 40.74931107467757, -73.9868422775547 40.74924506042546, -73.98676317970046 40.749212046914415, -73.9866840902161 40.749179018041396, -73.98654170588975 40.74911960140967, -73.98614617388499 40.74895458539953, -73.98480133968964 40.74839358146502)))')



# coding=utf-8
""" Contains objects related to cursors.

"""

from src.oogeo.objects.selectionset import SelectionSet
from src.oogeo.functions.database import convert_feature_to_row


class SearchCursor(SelectionSet):
    """ Provides iterable for walking through table query results

    """

    def __init__(self, table, field_names, query_string, sql_clause, allow_invalid=False):
        """ Initializes the search cursor object.

        :param table: The table that the search cursor is based on.
        :param field_names: The list of field names to return in the cursor.
        :param query_string: The query string to define the contents of the cursor.
        :param sql_clause: Any SQL pre/post clauses to be applied.
        :param allow_invalid: If true, errors for invalid geometries will be ignored. Set this
            flag to true when loading datasets where you expect to encounter invalid geometries.
        """
        super(SearchCursor, self).__init__(table, field_names, query_string, sql_clause, allow_invalid)
        self.current_feature = None

    def __iter__(self):
        return self

    def __next__(self):
        """ Iterates over the rows in the cursor.

        :return: The next row in the cursor as a list of field values.
        """
        self.current_feature = next(self.results)

        if self.current_feature is None:
            raise StopIteration

        if '*' in self.field_names:
            self.field_names = self.source_table.field_names + [self.source_table.oid]
            self.field_names = self.field_names + [self.source_table.shape] if hasattr(self.source_table, 'shape') else self.field_names

        return convert_feature_to_row(self.source_table, self.current_feature, self.field_names, self.allow_invalid)

    def __exit__(self, exc_type, exc_value, traceback):
        """ Destroys the results layer after iteration is complete.

        """
        self.current_feature = None
        super(SearchCursor, self).__exit__(exc_type, exc_value, traceback)

    def __len__(self):
        """ Returns the number of features in the cursor.

        :return: The number of features in the cursor
        """
        return self.results.rowcount


class InsertCursor(object):
    """ Provides cursor for inserting rows into a table.

    """

    def __init__(self, layer, field_names):
        """ Initializes the `InsertCursor()` object.

        :param layer: The layer that the cursor will insert rows into.
        :param field_names: The field names to be altered.
        """
        self.layer = layer

        # validate field names
        if isinstance(field_names, str):
            field_names = layer.field_names if field_names == '*' else [field_names]
        self.field_names_to_write = field_names

    def insert_row(self, row):
        """ Inserts a row into the table.

        :param row: The row to be inserted.
        """

        # map field names to values
        insert_fields = {}
        geometry_columns = [column.lower() for column in self.layer.get_geometry_field_names(include_hidden=True)] if hasattr(self.layer, 'get_geometry_field_names') else []
        for field_name, value in row.items():
            if field_name.lower() in geometry_columns + ['shape@']:
                field_name = self.layer.shape if field_name.lower() == 'shape@' else field_name
                insert_fields[field_name] = value.to_postgis() if hasattr(value, 'to_postgis') else value
            elif field_name.lower() == 'oid@' or field_name.lower() == self.layer.oid.lower():
                continue
            elif field_name in self.layer.field_names:
                insert_fields[field_name] = value
            else:
                raise KeyError('Could not find field {}'.format(field_name))

        insert_statement = self.layer.alchemy_table.insert().values(**insert_fields)
        self.layer.workspace.execute_simple_sql(insert_statement)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return


class UpdateCursor(SearchCursor):
    """ Provides a cursor for updating rows in the database.

    """
    def __init__(self, table, field_names, query_string, sql_clause, allow_invalid=False):
        """ Initializes the update cursor object.

        :param table: The table that the search cursor is based on.
        :param field_names: The list of field names to return in the cursor.
        :param query_string: The query string to define the contents of the cursor.
        :param sql_clause: Any SQL pre/post clauses to be applied.
        :param allow_invalid: If true, errors for invalid geometries will be ignored. Set this
            flag to true when loading datasets where you expect to encounter invalid geometries.
        """
        super(UpdateCursor, self).__init__(table, field_names, query_string, sql_clause, allow_invalid)

    def update_row(self, row):
        """ Updates the database with the values in the specified row.

        :param row: The row object to be updated.
        """
        row_id = self.current_feature['table_oid']
        update_fields = {}
        for field_name, value in row.items():
            if field_name.lower() == 'shape@' or \
                    (hasattr(self.source_table, 'shape') and field_name.lower() == self.source_table.shape.lower()):
                update_fields[self.source_table.shape] = value.to_postgis()
            elif 'shape@' not in field_name.lower() and 'oid@' not in field_name.lower() and field_name.lower() != self.source_table.oid.lower():
                update_fields[field_name] = value

        self.source_table.workspace.execute_simple_sql(
            self.source_table.alchemy_table.update().where(getattr(self.source_table.alchemy_table.c, self.source_table.oid) == row_id).values(**update_fields)
        )

    def delete_row(self):
        """ Deletes the current row.

        :return: None
        """
        row_id = self.current_feature['table_oid']
        self.source_table.workspace.execute_simple_sql(
            self.source_table.alchemy_table.delete().where(getattr(self.source_table.alchemy_table.c, self.source_table.oid) == row_id)
        )



# coding=utf-8
""" Contains objects related to workspaces.

Classes in the workspace file help manage the folders
where the application inputs and outputs are located.
They also help abstract the gis processing library
(arcpy) from the application processing logic.
"""

# import logging

from copy import deepcopy
from os import path
from sqlalchemy import create_engine, Table as AlchemyTable, Column, MetaData, Integer
from sqlalchemy.engine.reflection import Inspector

from src.oogeo.functions.constants import METADATA_TABLES
from src.oogeo.functions.conversions import upgrade_shapely_geometry
from src.oogeo.functions.database import create_in_clause
from src.oogeo.objects.layer import Layer
from src.oogeo.objects.table import Table
from geoalchemy2 import Geometry, WKTElement


def create_database_functions(workspace):
    """ Creates required functions in the specified workspace.

    :param workspace: The workspace object connected to the database where the functions will be created.
    :return: None
    """

    sql = "CREATE OR REPLACE FUNCTION shares_a_segment(analysis_geometry geometry, edge_buffer_geometry geometry) " \
          "RETURNS boolean AS $$ " \
          "DECLARE " \
          "shares_a_segment boolean := FALSE; " \
          "point RECORD; " \
          "last_point RECORD; " \
          "line_segment geometry; " \
          "point_cursor CURSOR(analysis_geometry geometry) FOR SELECT temp_table.geom FROM (SELECT (ST_Dump(ST_Points(analysis_geometry))).geom AS geom) AS temp_table; " \
          "BEGIN " \
          "IF ST_Disjoint(analysis_geometry, edge_buffer_geometry) THEN " \
          "RETURN FALSE; " \
          "END IF; " \
          "OPEN point_cursor(analysis_geometry); " \
          "LOOP " \
          "FETCH point_cursor INTO point; " \
          "EXIT WHEN NOT FOUND; " \
          "IF last_point IS NULL THEN " \
          "last_point := point; " \
          "CONTINUE; " \
          "END IF; " \
          "line_segment := ST_MakeLine(last_point.geom, point.geom); " \
          "IF ST_Within(line_segment, edge_buffer_geometry) THEN " \
          "shares_a_segment := TRUE; " \
          "EXIT; " \
          "END IF; " \
          "last_point := point; " \
          "END LOOP; " \
          "CLOSE point_cursor; " \
          "RETURN shares_a_segment; " \
          "END; $$ " \
          "LANGUAGE PLPGSQL;"
    workspace.execute_simple_sql(sql, use_transaction=False)


class Workspace(object):
    """ Controls the workspace used by the application

    The workspace class tracks the input and output directories
    used by the application.
    """

    def __init__(self, connection_info, schema='public'):
        """ Initializes the workspace object.

        """
        assert connection_info, 'A database connection must be specified.'
        assert 'user' in connection_info, 'A database username must be specified.'
        assert 'password' in connection_info, 'A database password must be specified.'
        assert 'host' in connection_info, 'A database host must be specified (use "localhost" for local geodatabases).'
        assert 'port' in connection_info, 'A database port must be specified (default is 5432).'
        self.alchemy_engine = None
        self.connection_info = deepcopy(connection_info)
        self.connect()

        self.database_name = connection_info['database'] if 'database' in connection_info else None
        self.pandas_connection = None
        self.is_gdb = True
        self.schema = schema

    def open_pandas_connection(self):
        """

        :return:
        """
        self.connect()
        self.pandas_connection = self.alchemy_engine

    def close_pandas_connection(self):
        """

        :return:
        """
        if self.pandas_connection is None:
            return
        self.pandas_connection = None

    def delete_feature_classes(self, feature_classes, schema=None):
        """ Deletes feature classes in the geodatabase without requiring a `Layer()` object.

        :param feature_classes: The list of feature classes to be deleted.
        :return: None
        """

        if isinstance(feature_classes, str):
            feature_classes = [feature_classes]

        if not schema:
            schema = self.schema
        for feature_class in feature_classes:
            self.execute_simple_sql(f'DROP TABLE IF EXISTS "{schema}"."{feature_class}"', use_transaction=False)
            self.execute_simple_sql(f'DROP SEQUENCE IF EXISTS "{schema}"."{feature_class}_seq"', use_transaction=False)
            self.execute_simple_sql(f'DROP FUNCTION IF EXISTS "{schema}"."edge_buffers_{feature_class}"', use_transaction=False)

    def import_shapefile(self, shapefile_path, layer_name,
        projection=None, schema=None):
        """ Imports the specified shapefile into the geodatabase as the specified layer name

        :param shapefile_path: The fully-qualified path to the shapefile to be imported.
        :param layer_name: The name of the layer in the geodatabase.
        :param projection: If a WKID is provided, the shapefile will be projected into the
            specified projection before importing.
        :return: A new `Layer()` object representing the new layer in the geodatabase.
        """

        try:
            import geopandas
        except Exception as e:
            raise ValueError('Geopandas is not installed')

        assert self.is_gdb, 'Shapefiles can only be imported into geodatabase workspaces.'
        assert path.isfile(shapefile_path), 'A shapefile must be specified.'

        # logging.info('Importing layer %s from %s', layer_name, shapefile_path)

        # load the shapefile and get it's projection
        # logging.info('Loading the shapefile into geopandas')
        shapefile_gdf = geopandas.read_file(shapefile_path)
        epsg = int(shapefile_gdf.crs.to_epsg())

        # project the shapefile if necessary
        if projection and projection != epsg:
            shapefile_gdf.to_crs(epsg=projection)
            epsg = projection

        if not schema:
            schema = self.schema

        # upgrade geometries to multis
        shapefile_gdf.geometry = shapefile_gdf.geometry.apply(upgrade_shapely_geometry)

        # get the layer's geometry type
        geometry_type = shapefile_gdf.geometry.geom_type[0].upper()

        # translate the geometries to geoalchemy (makes them storable in the db)
        if epsg == 102718:
            epsg = 2263
        shapefile_gdf['geom'] = shapefile_gdf.geometry.apply(lambda x: WKTElement(x.wkt, srid=epsg))
        shapefile_gdf.drop(shapefile_gdf.geometry.name, 1, inplace=True)

        # save the file into the geodatabase
        shapefile_gdf.to_sql(layer_name, self.alchemy_engine, if_exists='fail', index_label='id', schema=schema,
            dtype={'geom': Geometry(geometry_type, srid=epsg)})

        # set the object id as the primary key
        self.execute_simple_sql(f'ALTER TABLE "{schema}"."{layer_name}" ADD PRIMARY KEY ("id");', use_transaction=False)

        # create a sequence for the table
        self.create_sequence(layer_name, 'id', schema=schema)

        return Layer(layer_name, self)

    def create_sequence(self, table_name, oid_field, schema=None):
        """

        :param table_name:
        :param oid_field:
        :return:
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        # get the highest oid value and use it to create a sequence for the table
        result = self.execute_simple_sql(f'SELECT MAX("{oid_field}") as max_oid FROM "{schema}"."{table_name}"')
        row = next(result)
        start_oid = int(row['max_oid']) + 1 if row is not None and row['max_oid'] is not None else 1
        sequence_name = f'"{schema}"."{table_name}_{oid_field}_seq"'
        sql = f'CREATE SEQUENCE IF NOT EXISTS {sequence_name} START WITH {start_oid}'
        self.execute_simple_sql(sql)
        self.execute_simple_sql(f'ALTER TABLE "{schema}"."{table_name}" ALTER COLUMN "{oid_field}" SET DEFAULT nextval(\'{sequence_name}\');', use_transaction=False)
        self.execute_simple_sql(f'ALTER SEQUENCE {sequence_name} OWNED BY "{schema}"."{table_name}"."{oid_field}"', use_transaction=False)

        return sequence_name

    def list_feature_classes(self, wildcards=None, feature_type='All', schema=None):
        """ Lists the feature classes in the workspace, limited by name and feature type.

        :param wildcards: String filter on the results returned. If no wildcard is specified, all values are returned.
        :param feature_type: Feature type to be returned. Valid options are: Annotation, Arc, Dimension, Edge,
            Junction, Label, Line, Multipatch, Node, Point, Polygon, Polyline, Region, Route, Tic, All (default).
        :return: If input_path is a geodatabase, a list of feature class names; if shapefiles, a list of file paths.
        """

        # validate wildcards param
        if not wildcards:
            wildcards = []
        if isinstance(wildcards, str):
            wildcards = [wildcards]

        # set the proper schema
        if not schema:
            schema = self.schema


        # create basic where clause
        where_clause = f"WHERE f_table_schema = '{schema}'"
        # set up wildcard search if necessary
        for wildcard in wildcards:
            wildcard = wildcard.replace('*', '%%')
            where_clause = f"{where_clause} AND f_table_name LIKE '{wildcard}'"

        with self.alchemy_engine.connect() as connection:
            sql = f'SELECT DISTINCT f_table_name FROM geometry_columns {where_clause};'
            # logging.debug('List feature classes sql: %s', sql)
            table_names = connection.execute(sql)
        return [row['f_table_name'] for row in table_names]

    def list_tables(self, wildcards=None, schema=None):
        """ Lists the tables in the workspace, limited by name and table type.

        :param wildcard: String filter on the results returned. If no wildcard is specified, all values are returned.
        :param table_type: Feature type to be returned. Valid options are: dBASE, INFO, All (default).
        :return: A list of table names.
        """
        # validate wildcards param
        if not wildcards:
            wildcards = []
        if isinstance(wildcards, str):
            wildcards = [wildcards]

        # set the proper schema
        if not schema:
            schema = self.schema

        # create basic where clause
        not_in_clause = create_in_clause(METADATA_TABLES, use_quotes=True, field_name='tablename', not_operator=True)
        where_clause = f"WHERE schemaname = '{schema}' AND {not_in_clause}"
        # set up wildcard search if necessary
        for wildcard in wildcards:
            wildcard = wildcard.replace('*', '%%')
            where_clause = f"{where_clause} AND tablename LIKE '{wildcard}'"

        with self.alchemy_engine.connect() as connection:
            sql = "SELECT DISTINCT tablename FROM pg_catalog.pg_tables {};".format(where_clause)
            table_name_results = connection.execute(sql)

        table_names = [row['tablename'] for row in table_name_results]
        feature_classes = self.list_feature_classes(schema=schema)
        return [row for row in filter(lambda x: x not in feature_classes, table_names)]

    def load(self, table, schema=None):
        """

        """
        if table in self.list_tables(schema=schema):
            return Table(table, self, schema=schema)
        if table in self.list_feature_classes(schema=schema):
            return Layer(table, self, schema=schema)
        raise ValueError('Could not load specified table.')

    def list_tables_and_layers(self, schema=None):
        """ List both the tables and layers in a geodatabase.

        """
        # set the proper schema
        if not schema:
            schema = self.schema

        input_features = self.list_feature_classes(schema=schema)
        input_tables = self.list_tables(schema=schema)
        return input_features + input_tables

    def create_spatial_index(self, feature_class_name, shape_field_name, schema=None):
        """ Creates a spatial index on the specified shape field.

        :param feature_class_name: The name of the feature class to be altered.
        :param shape_field_name: The name of the shape field to be indexed.
        :return: None
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        spatial_index_name = f'idx_{feature_class_name}_{shape_field_name}'
        self.execute_simple_sql(f'DROP INDEX IF EXISTS "{schema}"."{spatial_index_name}";')
        self.execute_simple_sql(f'CREATE INDEX "{spatial_index_name}" ON "{schema}"."{feature_class_name}" USING GIST ("{shape_field_name}");', use_transaction=False)

    def create_index(self, table_name, field_name, schema=None):
        """ Creates an index on a field

        :param table_name: The name of the table to be modified.
        :param field_name: The name of the field to be indexed.
        :return: None
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        index_name = 'idx_{0}_{1}'.format(table_name, field_name)
        self.execute_simple_sql(f'CREATE INDEX IF NOT EXISTS "{index_name}" ON "{schema}".{table_name} ("{field_name}");', use_transaction=False)

    def list_indexes(self, table_name, schema=None):
        """ Lists the names of the indexes on the specified table

        :param table_name: The name of the table to be analyzed.
        :return: A list of index names assofiated with the table.
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        with self.alchemy_engine.connect() as connection:
            index_names = connection.execute(
                f"SELECT indexname FROM pg_indexes WHERE schemaname = '{schema}' AND tablename = '{table_name}';")
        return [row['indexname'] for row in index_names]

    def create_geometry_column(self, table_name, field_name, field_type, spatial_reference=2263, schema=None, three_d=False):
        """ Adds a geometry column to the specified table.

        """
        # standardize geometry type
        geometry_type = field_type.upper()
        if geometry_type == 'POLYGON':
            geometry_type = 'MULTIPOLYGON'
        elif geometry_type == 'POLYLINE' or geometry_type == 'LINESTRING':
            geometry_type = 'MULTILINESTRING'

        # set the proper schema
        if not schema:
            schema = self.schema

        # set 2D/3D
        dimensions = '2' if not three_d else '3'

        # create the geometry column
        self.execute_simple_sql(
            f"SELECT AddGeometryColumn('{schema}', '{table_name}', '{field_name}', {spatial_reference}, '{geometry_type}', {dimensions});", use_transaction=False)

    def has_geometry_search_column(self, feature_class_name, schema=None):
        """ Indicates if a feature class has an 'edge_buffers' column (used for querying edges).

        :param feature_class_name: The name of the feature class to be analyzed.
        :return: True if the table has a geometry search column, otherwise False.
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        with self.alchemy_engine.connect() as connection:
            column_infos = connection.execute(
                f"SELECT f_geometry_column as name, type, srid FROM geometry_columns WHERE f_table_schema = '{schema}' AND f_table_name = '{feature_class_name}' AND LOWER(f_geometry_column) = 'edge_buffers';")
            if column_infos.rowcount > 0:
                return True
        return False

    def drop_geometry_search_column(self, feature_class_name, schema=None):
        """

        :param feature_class_name:
        :return:
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        # do not create if column already exists
        if not self.has_geometry_search_column(feature_class_name, schema=schema):
            return

        # drop column if it exists
        trigger_name = 'edge_buffers_{}'.format(feature_class_name)
        ucase_trigger_name = 'EDGE_BUFFERS_{}'.format(feature_class_name)

        # drop the trigger itself
        self.execute_simple_sql(f'DROP TRIGGER IF EXISTS "{trigger_name}" ON "{schema}"."{feature_class_name}"', use_transaction=False)
        self.execute_simple_sql(f'ALTER TABLE "{schema}"."{feature_class_name}" DROP COLUMN IF EXISTS "edge_buffers"')

        self.execute_simple_sql(f'DROP TRIGGER IF EXISTS "{ucase_trigger_name}" ON "{schema}"."{feature_class_name}"', use_transaction=False)
        self.execute_simple_sql(f'ALTER TABLE "{schema}"."{feature_class_name}" DROP COLUMN IF EXISTS "EDGE_BUFFERS"')

    def create_geometry_search_column(self, feature_class_name, geometry_type, shape_field_name, spatial_reference=2263, schema=None):
        """ Creates an 'edge_buffers' column on a feature class (used for querying edges).

        :param feature_class_name: The feature class to be modified.
        :return: None
        """
        # standardize geometry type
        geometry_type = geometry_type.upper()
        if geometry_type == 'POLYGON':
            geometry_type = 'MULTIPOLYGON'
        elif geometry_type == 'POLYLINE' or geometry_type == 'LINESTRING':
            geometry_type = 'MULTILINESTRING'

        # do not create search columns on point layers
        if geometry_type == 'POINT':
            return

        # set the proper schema
        if not schema:
            schema = self.schema

        # do not create if column already exists
        if self.has_geometry_search_column(feature_class_name, schema=schema):
            return

        # drop column if it exists
        self.execute_simple_sql(f'ALTER TABLE "{schema}"."{feature_class_name}" DROP COLUMN IF EXISTS "edge_buffers"')

        # create the edge_buffers geometry column
        self.create_geometry_column(feature_class_name, 'edge_buffers', 'MULTIPOLYGON', spatial_reference, schema=schema)

        # populate the edge_buffers column for existing rows
        spatial_sql = 'ST_Boundary("{}")'.format(shape_field_name) if geometry_type == 'MULTIPOLYGON' else '"{}"'.format(shape_field_name)
        spatial_sql = 'ST_Multi(ST_Buffer({}, 0.001, \'endcap=square join=mitre\'))'.format(spatial_sql)
        self.execute_simple_sql(f'UPDATE "{schema}"."{feature_class_name}" SET "edge_buffers" = {spatial_sql}')

        # add a spatial index to the column
        self.create_spatial_index(feature_class_name, 'edge_buffers', schema=schema)

        # create the trigger function to populate the edge buffers column
        trigger_name = 'edge_buffers_{}'.format(feature_class_name)
        if geometry_type == 'MULTILINESTRING':
            self.execute_simple_sql('CREATE OR REPLACE FUNCTION "{schema}"."{trigger_name}"() RETURNS trigger AS ${trigger_name}$ '
                                    'BEGIN '
                                    'NEW."edge_buffers" := ST_Multi(ST_Buffer(NEW."{shape_field_name}", 0.001, \'endcap=square join=mitre\')); '
                                    'RETURN NEW; '
                                    'END; ${trigger_name}$ LANGUAGE plpgsql;'.format(trigger_name=trigger_name, shape_field_name=shape_field_name, schema=schema), use_transaction=False)

        if geometry_type == 'MULTIPOLYGON':
            self.execute_simple_sql('CREATE OR REPLACE FUNCTION "{schema}"."{trigger_name}"() RETURNS trigger AS ${trigger_name}$ '
                                    'BEGIN '
                                    'NEW."edge_buffers" := ST_Multi(ST_Buffer(ST_Boundary(NEW."{shape_field_name}"), 0.001, \'endcap=square join=mitre\')); '
                                    'RETURN NEW; '
                                    'END; ${trigger_name}$ LANGUAGE plpgsql;'.format(trigger_name=trigger_name, shape_field_name=shape_field_name, schema=schema), use_transaction=False)

        # create the trigger itself
        self.execute_simple_sql('CREATE TRIGGER "{trigger_name}" BEFORE INSERT OR UPDATE ON "{schema}"."{feature_class_name}" '
                                'FOR EACH ROW EXECUTE PROCEDURE {schema}.{trigger_name}();'.format(trigger_name=trigger_name, feature_class_name=feature_class_name, schema=schema), use_transaction=False)

    def has_centroid_column(self, feature_class_name, schema=None):
        """ Indicates if a feature class has an 'centroid' column (used for querying).

        :param feature_class_name: The name of the feature class to be analyzed.
        :return: True if the table has a centroid column, otherwise False.
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        with self.alchemy_engine.connect() as connection:
            column_infos = connection.execute(
                "SELECT f_geometry_column as name, type, srid FROM public.geometry_columns "
                "WHERE f_table_schema = '{}' AND f_table_name = '{}' AND LOWER(f_geometry_column) = 'centroid';".format(schema, feature_class_name))
            if column_infos.rowcount > 0:
                return True
        return False

    def drop_centroid_column(self, feature_class_name, schema=None):
        """

        :param feature_class_name:
        :return:
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        # do not create if column already exists
        if not self.has_centroid_column(feature_class_name, schema=schema):
            return

        # drop column if it exists
        self.execute_simple_sql(f'ALTER TABLE "{schema}"."{feature_class_name}" DROP COLUMN IF EXISTS "centroid"')
        self.execute_simple_sql(f'ALTER TABLE "{schema}"."{feature_class_name}" DROP COLUMN IF EXISTS "CENTROID"')

    def create_centroid_columm(self, feature_class_name, shape_field_name, spatial_reference=2263, inside=False, schema=None):
        """

        :param feature_class_name:
        :param shape_field_name:
        :param spatial_reference:
        :return:
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        # do not create if column already exists
        create = not self.has_centroid_column(feature_class_name, schema=schema)
        if create:
            # create the centroid geometry column
            self.create_geometry_column(feature_class_name, 'centroid', 'POINT', spatial_reference, schema=schema)

        # populate the centroid column for existing rows
        if not inside:
            spatial_sql = f'ST_Centroid("{shape_field_name}")'
        else:
            spatial_sql = f'ST_PointOnSurface("{shape_field_name}")'
        self.execute_simple_sql(f'UPDATE "{schema}"."{feature_class_name}" SET "centroid" = {spatial_sql}')

        # create an index on the centroid column
        if create:
            self.create_spatial_index(feature_class_name, 'centroid', schema=schema)
        else:
            self.vacuum(feature_class_name, schema=schema)


    def has_midpoint_column(self, feature_class_name, schema=None):
        """ Indicates if a feature class has an 'midpoint' column (used for querying).

        :param feature_class_name: The name of the feature class to be analyzed.
        :return: True if the table has a midpoint column, otherwise False.
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        with self.alchemy_engine.connect() as connection:
            column_infos = connection.execute(
                "SELECT f_geometry_column as name, type, srid FROM public.geometry_columns "
                "WHERE f_table_schema = '{}' AND f_table_name = '{}' AND LOWER(f_geometry_column) = 'midpoint';".format(schema, feature_class_name))
            if column_infos.rowcount > 0:
                return True
        return False

    def drop_midpoint_column(self, feature_class_name, schema=None):
        """ Drops the `midpoint` column from the specified feature class.

        :param feature_class_name: The name of the feature class to be altered.
        :return: None
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        # do not create if column already exists
        if not self.has_midpoint_column(feature_class_name, schema=schema):
            return

        # drop column if it exists
        self.execute_simple_sql(f'ALTER TABLE "{schema}"."{feature_class_name}" DROP COLUMN IF EXISTS "midpoint"')
        self.execute_simple_sql(f'ALTER TABLE "{schema}"."{feature_class_name}" DROP COLUMN IF EXISTS "MIDPOINT"')

    def create_midpoint_columm(self, feature_class_name, shape_field_name, spatial_reference=2263, schema=None):
        """

        :param feature_class_name:
        :param shape_field_name:
        :param spatial_reference:
        :return:
        """

        # set the proper schema
        if not schema:
            schema = self.schema

        # do not create if column already exists
        if self.has_midpoint_column(feature_class_name, schema=schema):
            return

        # create the midpoint geometry column
        self.create_geometry_column(feature_class_name, 'midpoint', 'POINT', spatial_reference, schema=schema)

        # populate the midpoint column for existing rows
        spatial_sql = f'ST_LineInterpolatePoint(ST_LineMerge("{shape_field_name}"), 0.5)'
        self.execute_simple_sql(f'UPDATE "{schema}"."{feature_class_name}" SET "midpoint" = {spatial_sql}')

        # create a spatial index on the midpoint column
        self.create_spatial_index(feature_class_name, 'midpoint', schema=schema)

    def unlog_table(self, table_name, schema=None):
        """ Sets a table to be "unlogged", which will make the table perform faster, but data
        will not be recoverable is the database crashes.

        :param table_name: The name of the table to be set to "unlogged".
        :return: None
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        sql = f'ALTER TABLE "{schema}"."{table_name}" SET UNLOGGED;'
        self.execute_simple_sql(sql, use_transaction=False)

    def vacuum(self, table_name=None, analyze=True, schema=None):
        """ Cleans up any data files produced by the table. Can be used to manage hard-drive space when
        doing large update operations.

        :param table_name: The table to be vacuumed.
        :param analyze: If True, the `ANALYZE` command will also be issues to reanalyze the statistics
            for the table or database.
        :return: None
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        if table_name:
            if analyze:
                sql = f'VACUUM ANALYZE "{schema}"."{table_name}";'
            else:
                sql = f'VACUUM "{schema}"."{table_name}";'
        else:
            if analyze:
                sql = 'VACUUM ANALYZE;'
            else:
                sql = 'VACUUM;'
        self.execute_simple_sql(sql, use_transaction=False)

    def create_feature_class(self, feature_class_name, geometry_type, spatial_reference=2263, schema=None):
        """ Creates a new feature class in the workspace.

        :param feature_class_name: The name of the feature class to be created.
        :param geometry_type: The geometry type of the feature class. Valid values are:
            "POINT", "MULTIPOINT", "POLYGON", and "POLYLINE"
        :param spatial_reference: The WKID of the spatial reference for the new feature class. The default
            is 2263 (New York State Plane)
        :return: A `Layer` object representing the new feature class.
        """
        geometry_type = geometry_type.upper()
        if geometry_type == 'POLYGON':
            geometry_type = 'MULTIPOLYGON'
        elif geometry_type == 'POLYLINE' or geometry_type == 'LINESTRING':
            geometry_type = 'MULTILINESTRING'

        # set the proper schema
        if not schema:
            schema = self.schema
        self.create_schema(schema)

        alchemy_table = AlchemyTable(feature_class_name, MetaData(),
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('geom', Geometry(geometry_type, srid=spatial_reference)),
                                    schema=schema
        )

        alchemy_table.create(self.alchemy_engine)

        # add spatial index
        self.create_spatial_index(feature_class_name, 'geom', schema=schema)

        return Layer(alchemy_table, self, schema=schema)

    def create_table(self, table_name, schema=None):
        """ Creates a new table in the geodatabase.

        :param table_name: The name of the new table
        :return: A `Table()` object representing the new table.
        """
        # set the proper schema
        if not schema:
            schema = self.schema
        self.create_schema(schema)

        # create a new table with a primary key
        alchemy_table = AlchemyTable(table_name, MetaData(),
            Column('id', Integer, primary_key=True, autoincrement=True),
                                     schema=schema
        )
        alchemy_table.create(self.alchemy_engine)

        return Table(alchemy_table, self, schema=schema)

    def create_view(self, view_name, view_query, schema=None):
        """ Creates a view based on the specified sql query.

        :param view_name: The name to be given to the view.
        :param view_query: The SQL `SELECT` statement to be used to create the view.
        """
        # set the proper schema
        if not schema:
            schema = self.schema
        self.create_schema(schema)

        sql = f'CREATE OR REPLACE VIEW "{schema}"."{view_name}" AS {view_query};'
        # logging.debug('Create view sql: {}'.format(sql))
        self.execute_simple_sql(sql)

        sql = f'ALTER TABLE "{schema}"."{view_name}" OWNER TO postgres;'
        self.execute_simple_sql(sql)

    def drop_view(self, view_name, schema=None, cascade=False):
        """ Drops a view if it exists in the database.

        :param view_name: The name of the view to drop.
        :param cascade: If `True` a cascading drop will be performed.
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        sql = f'DROP VIEW IF EXISTS "{schema}"."{view_name}"'
        if cascade:
            sql = f'{sql} CASCADE'
        sql = f'{sql};'
        # logging.debug('Drop view sql: {}'.format(sql))
        self.execute_simple_sql(sql)

    def drop_table(self, table_name, schema=None, cascade=False):
        """

        :param table_name:
        :param cascade: If `True` a cascading drop will be performed.
        :return:
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        if self.has_geometry_search_column(table_name, schema=schema):
            self.drop_geometry_search_column(table_name, schema=schema)
        sql = f'DROP TABLE IF EXISTS "{schema}"."{table_name}"'
        if cascade:
            sql = f'{sql} CASCADE'
        sql = f'{sql};'

        self.execute_simple_sql(sql)

    def has_schema(self, schema):
        """ Indicates if the database has the specified schema.

        :param schema: The name of the schema to be checked.
        :return: `True` if the specified schema exists, else `False`.
        """
        sql = f"SELECT COUNT(*) AS \"count\" FROM information_schema.schemata WHERE schema_name = '{schema}';"
        scalar = self.execute_simple_sql(sql, use_transaction=False)
        row = next(scalar)
        return True if int(row['count']) > 0 else False

    def list_schemas(self, include_hidden=False):
        """ List the names of the schemas in the workspace.

        :param include_hidden: If `True`, the list will include schemas usually hidden from view, such as
            "pg_catalog" and "information_schema".
        :return: A list of schema names currently in the workspace.
        """
        sql = "SELECT schema_name FROM information_schema.schemata"
        if not include_hidden:
            sql = f"{sql} WHERE schema_name NOT IN ('pg_toast', 'pg_catalog', 'information_schema')"
        with self.alchemy_engine.connect() as connection:
            schema_names = connection.execute(sql)
        return [row['schema_name'] for row in schema_names]


    def create_schema(self, schema):
        """ Creates a new schema in the workspace.

        :param schema: The schema name to be created.
        :return: A `Workspace` object with the default schema set to the new schema name.
        """
        sql = f'CREATE SCHEMA IF NOT EXISTS {schema};'
        self.execute_simple_sql(sql, use_transaction=False)
        return Workspace(self.connection_info, schema=schema)

    def drop_schema(self, schema, cascade=False):
        """ Drops the specified schema.

        :param schema: The schema name to be dropped.
        :param cascade: If `True` a cascading drop will be performed.
        :return: None.
        """
        sql = f'DROP SCHEMA IF EXISTS {schema}'
        if cascade:
            sql = f'{sql} CASCADE'
        sql = f'{sql};'
        self.execute_simple_sql(sql, use_transaction=False)

    def get_geometry_columns(self, table_name, include_hidden=False, schema=None):
        """ Finds the geometry column for the specified table.

        :param table_name: The table to find.
        :param include_hidden: If true, columns that are otherwise hidden such as edge_buffers, midpoint, and centroid
            will be included in the results.
        :return: A dict representing the attributes of the geometry column ('name', 'type', and 'srid') or None
            if the table does not have a geometry column.
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        # find the name of the table's geometry column
        with self.alchemy_engine.connect() as connection:
            sql = "SELECT f_geometry_column as name, type, srid FROM public.geometry_columns " \
                  "WHERE f_table_schema = '{}' AND f_table_name = '{}'".format(schema, table_name)
            if not include_hidden:
                sql = '{} {}'.format(sql, "AND LOWER(f_geometry_column) NOT IN ('edge_buffers', 'centroid', 'midpoint')")
            sql = f'{sql};'

            column_infos = connection.execute(sql)
            if column_infos.rowcount == 0:
                return None

            return [info for info in column_infos]

    def get_oid_column(self, table_name, schema=None):
        """

        :param table_name:
        :return:
        """
        # set the proper schema
        if not schema:
            schema = self.schema

        inspector = Inspector(self.alchemy_engine)
        primary_keys = inspector.get_pk_constraint(table_name, schema=schema)
        return primary_keys['constrained_columns'][0]

    def connect(self):
        """ Opens the SqlAlchemy connection to the geodatabase.

        :return: None
        """
        if self.alchemy_engine is not None:
            return
        connection_string = f"postgresql://{self.connection_info['user']}:{self.connection_info['password']}@{self.connection_info['host']}:{self.connection_info['port']}"
        if 'database' in self.connection_info and self.connection_info['database'] is not None:
            connection_string = f"{connection_string}/{self.connection_info['database']}"
        self.alchemy_engine = create_engine(connection_string)

    def disconnect(self):
        """ Closes the SqlAlchemy connection to the geodatabase.

        :return: None
        """
        if self.alchemy_engine is None:
            return
        self.alchemy_engine.dispose()
        self.alchemy_engine = None

    def list_databases(self):
        """ Returns a list of databases in the PostGIS instance.

        :return: A list of databases in the PostGIS instance.
        """
        with self.alchemy_engine.connect() as connection:
            database_names = connection.execute("SELECT datname FROM pg_catalog.pg_database WHERE datname NOT IN ('template0', 'template1');")
        return [row['datname'] for row in database_names]

    def execute_simple_sql(self, query_string, use_transaction=True):
        """ Executes a simple SQL statement and returns the results.

        :param query_string: The SQL statement to execute.
        :param use_transaction: If True the execution of the statement will be wrapped in a transaction. If
            False the statement will be executed outside of a transaction (required when altering the database
            schema).
        :return: The results of the database query.
        """
        if use_transaction:
            with self.alchemy_engine.connect() as connection:
                return connection.execute(query_string)

        # to execute without a transaction, get an open connection and
        # commit the current transaction before executing the actual sql statement
        connection = self.alchemy_engine.connect()
        connection.execute('COMMIT;')
        results = connection.execute(query_string)
        connection.close()
        return results

    @staticmethod
    def create_geodatabase(connection_info, database_name):
        """ Creates a file geodatabase, if it doesn't exist.

        :param file_path: File path to the folder that will contain the geodatabase.
        :param database_name: File name of the geodatabase (should end with .gdb).
        :return: A workspace object that is connected to the new geodatabase.
        """
        # logging.info('Creating geodatabase %s', database_name)
        connection_info = {
            'host': connection_info['host'],
            'port': connection_info['port'],
            'user': connection_info['user'],
            'password': connection_info['password']
        }
        workspace = Workspace(connection_info)
        sql = "CREATE DATABASE {} WITH  OWNER = postgres ENCODING = 'UTF8' " \
                           "TABLESPACE = pg_default CONNECTION LIMIT = -1;".format(database_name)
        workspace.execute_simple_sql(sql, use_transaction=False)
        workspace.disconnect()
        new_connection_info = connection_info.copy()
        new_connection_info['database'] = database_name
        new_workspace = Workspace(new_connection_info)
        new_workspace.execute_simple_sql('CREATE EXTENSION IF NOT EXISTS postgis', use_transaction=False)
        new_workspace.execute_simple_sql('CREATE EXTENSION IF NOT EXISTS dblink', use_transaction=False)
        create_database_functions(new_workspace)
        return new_workspace

    @staticmethod
    def drop_geodatabase(connection_info, database_name):
        """ Drops the specified database from the host server.

        :param connection_info: A dictionary containing database connection settings.
        :param database_name: The name of the database to be dropped.
        :return: None.
        """
        connection_info = {
            'host': connection_info['host'],
            'port': connection_info['port'],
            'user': connection_info['user'],
            'password': connection_info['password']
        }
        workspace = Workspace(connection_info)
        workspace.execute_simple_sql("DROP DATABASE IF EXISTS {}".format(database_name), use_transaction=False)



# coding=utf-8
""" This is a fake program used for testing the `Processor()` class's ability to launch external processes.

"""

import time

# if running from a command line, gather command line args and call process_frequency
if __name__ == '__main__':
    print('This is a test process that performs no functions. Ending in 5 seconds...')
    time.sleep(5)



"""
Unit tests for the supporting utility functions in the 'utils' directory. For more information see the
[API Tests](https://github.com/envelopecity/GIS/tree/dev#api-tests) section of the GIS repository.
"""

import unittest

from src.oogeo.functions.database import create_in_clause
from src.oogeo.functions.conversions import string_to_units, parse_expression


class TableMock(object):

    def __init__(self, table_name):
        self.table_name = table_name
        self.oid = 'id'
        self.shape = 'geom'

    def get_field_type(self, field_name):
        if field_name == 'string_field':
            return 'VARCHAR 100'
        return 'DOUBLE PRECISION'


class UtilsUnitTests(unittest.TestCase):

    def test_create_in_clause(self):
        in_clause = create_in_clause([1, 2])
        self.assertEqual(in_clause, ' IN (1,2)')

        in_clause = create_in_clause([1, 2], use_quotes=True)
        self.assertEqual(in_clause, " IN ('1','2')")

        in_clause = create_in_clause(['1', '2'])
        self.assertEqual(in_clause, ' IN (1,2)')

        in_clause = create_in_clause(['1', '2'], use_quotes=True)
        self.assertEqual(in_clause, " IN ('1','2')")

        in_clause = create_in_clause([1, 2], field_name='field')
        self.assertEqual(in_clause, '"field" IN (1,2)')

        in_clause = create_in_clause([1, 2], field_name='field', not_operator=True)
        self.assertEqual(in_clause, '"field" NOT IN (1,2)')

        in_clause = create_in_clause([1, 2], field_name='field', use_quotes=True)
        self.assertEqual(in_clause, "\"field\" IN ('1','2')")

        in_clause = create_in_clause([1, 2], field_name='field', use_quotes=True, not_operator=True)
        self.assertEqual(in_clause, "\"field\" NOT IN ('1','2')")

    def test_string_to_units(self):
        self.assertEqual(string_to_units('100 FEET', 'FEET'), 100)
        self.assertEqual(string_to_units('1 FOOT', 'FEET'), 1)
        self.assertEqual(string_to_units('100 METERS', 'METERS'), 100)
        self.assertEqual(string_to_units('1 METER', 'METERS'), 1)
        self.assertAlmostEqual(string_to_units('100 FEET', 'METERS'), 30.48, 2)
        self.assertAlmostEqual(string_to_units('1 FOOT', 'METERS'), 0.3048, 4)
        self.assertAlmostEqual(string_to_units('100 METERS', 'FEET'), 328.08, 2)
        self.assertAlmostEqual(string_to_units('1 METER', 'FEET'), 3.2808, 4)

    def test_parse_expression(self):
        table_mock = TableMock('TABLE_NAME')
        self.assertEqual(parse_expression(table_mock, 'int_field', 0), 0)
        self.assertEqual(parse_expression(table_mock, 'double_field', 1.1), 1.1)
        self.assertEqual(parse_expression(table_mock, 'double_field', 1.1), 1.1)
        self.assertEqual(parse_expression(table_mock, 'id', 'id = 1'), 'id = 1')
        self.assertEqual(parse_expression(table_mock, 'id', 'id != 1'), 'id != 1')
        self.assertEqual(parse_expression(table_mock, 'string_field', "'1'"), "'1'")
        self.assertEqual(parse_expression(table_mock, 'string_field', "'string value'"), "'string value'")
        self.assertEqual(parse_expression(table_mock, 'string_field', "1"), "'1'")
        self.assertEqual(parse_expression(table_mock, 'string_field', "string value"), "'string value'")
        self.assertEqual(parse_expression(table_mock, 'string_field', "string's value's"), "'string''s value''s'")
        self.assertEqual(parse_expression(table_mock, 'string_field', "'string's value's'"), "'string''s value''s'")
        self.assertEqual(parse_expression(table_mock, 'string_field', "!FIELD_NAME!"), '"FIELD_NAME"')
        self.assertEqual(parse_expression(table_mock, 'string_field', 'NULL'), 'NULL')
        self.assertEqual(parse_expression(table_mock, 'int_field', 'NULL'), 'NULL')
        self.assertEqual(parse_expression(table_mock, 'string_field', None), 'NULL')
        self.assertEqual(parse_expression(table_mock, 'int_field', None), 'NULL')
        self.assertEqual(parse_expression(table_mock, 'int_field', "!FIELD_NAME!"), '"FIELD_NAME"')
        self.assertEqual(parse_expression(table_mock, 'string_field', "!FIELD_NAME! = 'string_value'"), '"FIELD_NAME" = \'string_value\'')
        self.assertEqual(parse_expression(table_mock, 'int_field', "!FIELD_NAME! = 'string_value'"), '"FIELD_NAME" = \'string_value\'')
        self.assertEqual(parse_expression(table_mock, 'int_field', "!FIELD_NAME! = 1"), '"FIELD_NAME" = 1')
        self.assertEqual(parse_expression(table_mock, 'string_field', "!FIELD_NAME! = 1"), '"FIELD_NAME" = 1')
        self.assertEqual(parse_expression(table_mock, 'string_field', "!FIELD_NAME! != 1"), '"FIELD_NAME" != 1')
        self.assertEqual(parse_expression(table_mock, 'int_field', "!FIELD_NAME! != 1"), '"FIELD_NAME" != 1')
        self.assertEqual(parse_expression(table_mock, 'int_field', "SUM(!FIELD_NAME!) + 1"), 'SUM("FIELD_NAME") + 1')
        self.assertEqual(parse_expression(table_mock, 'string_field', "SUM(!FIELD_NAME!) + 1"), 'SUM("FIELD_NAME") + 1')
        self.assertEqual(parse_expression(table_mock, 'string_field', "!SHAPE.AREA@SQUAREFEET!"), 'ST_Area("geom")')
        self.assertEqual(parse_expression(table_mock, 'int_field', "!SHAPE.AREA@SQUAREFEET!"), 'ST_Area("geom")')


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(UtilsUnitTests)
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
Unit tests for the Point class. For more information see the
[API Tests](https://github.com/envelopecity/GIS/tree/dev#api-tests) section of the GIS repository.
"""

from shapely.wkt import loads as wkt_to_geom

import unittest

from src.oogeo.objects.point import Point
import src.oogeo.objects.line
import src.oogeo.objects.polygon


class PointClassUnitTests(unittest.TestCase):

    def test_initialize_tuple(self):
        # initialize the point using a tuple
        point = (1, 1)
        point = Point(point, 2263)
        self.assertEqual(point.x, 1)
        self.assertEqual(point.y, 1)

    def test_initialize_dict(self):
        # initialize the point using a dictionary
        point = {'x': 1, 'y': 1}
        point = Point(point, 2263)
        self.assertEqual(point.x, 1)
        self.assertEqual(point.y, 1)

    def test_initialize_wkt(self):
        # initialize the point with well-known text
        wkt = "POINT (1 1)"
        point = Point(wkt, 2263)
        self.assertEqual(point.x, 1)
        self.assertEqual(point.y, 1)
        self.assertEqual(point.wkt, wkt)

    def test_initialize_spatial_reference(self):
        # initialize the point with well-known text and a spatial reference
        wkt = "POINT (1 1)"
        test_point = Point(wkt, 2263)
        self.assertEqual(test_point.spatial_reference, 2263)

    def test_initialize_geometry(self):
        # initialize the point using a shapely geometry
        wkt = "POINT (1 1)"
        point = wkt_to_geom(wkt)
        point = Point(point, 2263)
        self.assertEqual(point.x, 1)
        self.assertEqual(point.y, 1)

    def test_wkt(self):
        # initialize the point using a well-known text
        wkt = "POINT (1 1)"
        test_point = Point(wkt, 2263)
        self.assertEqual(test_point.wkt, wkt)

    def test_xy_properties(self):
        # verify that the x and y properties return appropriate values
        point = {'x': 1, 'y': 2}
        test_point = Point(point, 2263)
        self.assertEqual(test_point.x, 1)
        self.assertEqual(test_point.y, 2)

    def test_distance_to(self):
        # create two point that are one unit apart
        first_point = Point({'x': 1, 'y': 1}, 2263)
        second_point = Point({'x': 1, 'y': 2}, 2263)
        self.assertEqual(first_point.distance_to(second_point), 1)

    def test_buffer(self):
        # create a buffer around a point
        buffer_point = Point({'x': 1, 'y': 1}, 2263)
        buffer = buffer_point.buffer(1)
        self.assertIsInstance(buffer, src.oogeo.objects.polygon.Polygon)
        self.assertAlmostEqual(buffer.area, 3.14, 2)

    def test_snap_to_vertex(self):
        # create a line and set of snap targets to snap the line to
        point = Point(
            {'x': 3, 'y': 0}, 2263
        )

        snap_square = src.oogeo.objects.polygon.Polygon([
            {'x': 0, 'y': 0},
            {'x': 0, 'y': 2},
            {'x': 2, 'y': 2},
            {'x': 2, 'y': 0},
        ], 2263)
        snap_line = src.oogeo.objects.line.Line([
            {'x': 0, 'y': 0},
            {'x': 2, 'y': 0}
        ], 2263)
        snap_point = Point(
            {'x': 2, 'y': 0}, 2263
        )

        result = point.snap_to_vertex(snap_point, 2)
        self.assertIsInstance(result, Point)
        self.assertEqual(result.x, 2)
        self.assertEqual(result.y, 0)

        result = point.snap_to_vertex(snap_line, 2)
        self.assertIsInstance(result, Point)
        self.assertEqual(result.x, 2)
        self.assertEqual(result.y, 0)

        result = point.snap_to_vertex(snap_square, 2)
        self.assertIsInstance(result, Point)
        self.assertEqual(result.x, 2)
        self.assertEqual(result.y, 0)

    def test_project(self):
        point = Point(
            {'x': -80, 'y': 50}, 4326
        )
        new_point = point.project(26917)
        self.assertAlmostEqual(new_point.x, 571666.4475041276, 6)
        self.assertAlmostEqual(new_point.y, 5539109.815175673, 6)

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(PointClassUnitTests)
    unittest.TextTestRunner(verbosity=2).run(suite)
